import os
import sys
import getpass
import logging
import platform
from collections import OrderedDict
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import Any

import questionary

from ..common import EncryptionConfig
from ..utils import AccountManager, KeyManager
from ..version import __package_name__


class MenuAction(Enum):
    CREATE = "create"
    READ = "read"
    UPDATE = "update"
    DELETE = "delete"
    PASSWORD = "password"
    LIST = "list"
    QUIT = "quit"


@dataclass
class UIStrings:
    menu_create = "Create Account"
    menu_read = "Read Account"
    menu_update = "Update Account"
    menu_delete = "Delete Account"
    menu_password = "Password Test"
    menu_list = "List All Accounts"
    menu_quit = "Quit"
    menu_prompt = "Please choose an action:"

    prompt_username = "Please enter the username: "
    prompt_password = "Please enter the password: "
    prompt_cookies = "Please enter the cookies: "
    prompt_old_username = "Please enter the old username: "
    prompt_new_username = "Please enter the new username (leave blank to not update): "
    prompt_new_password = "Please enter the new password (leave blank to not update): "
    prompt_new_cookies = "Please enter the new cookies (leave blank to not update): "
    prompt_password_test = "Enter password (leave blank to not update): "

    msg_account_not_found = "Account not found."
    msg_no_accounts = "No accounts available."
    msg_operation_canceled = "Operation canceled."
    msg_exit = "Exiting the account management of v2dl."
    msg_invalid_choice = "Invalid choice, please try again."

    confirm_delete = "Are you sure you want to delete the account {username}?"
    confirm_yes = "Confirm"
    CONFIRM_NO = "Cancel"

    LIST_FORMAT = "Email: {username}, Exceed quota: {quota}, Exceed time: {time}, Created At: {created_at}, \nCookies: {cookies}\n"


class AccountManagerCLI:
    action_map: dict[str, Callable[..., Any]] = {}

    def __init__(self, encrypt_config: EncryptionConfig):
        self.logger = logging.getLogger(__package_name__)
        self.logger.setLevel(logging.INFO)
        self.strings = UIStrings()
        self.key_manager = KeyManager(self.logger, encrypt_config)
        self.account_manager = AccountManager(self.logger, self.key_manager)
        key_pair = self.key_manager.load_keys()
        self.private_key, self.public_key = key_pair.private_key, key_pair.public_key

    def clean_terminal(self) -> None:
        os.system("cls" if os.name == "nt" else "clear")  # nosec

    def get_pass(self, prompt: str = "Password: ") -> str:
        if platform.system() == "Windows":
            return input(prompt)
        return getpass.getpass(prompt)

    def get_menu_choices(self) -> list[dict[str, str]]:
        return [
            {"name": self.strings.menu_create, "value": MenuAction.CREATE.value},
            {"name": self.strings.menu_read, "value": MenuAction.READ.value},
            {"name": self.strings.menu_update, "value": MenuAction.UPDATE.value},
            {"name": self.strings.menu_delete, "value": MenuAction.DELETE.value},
            {"name": self.strings.menu_password, "value": MenuAction.PASSWORD.value},
            {"name": self.strings.menu_list, "value": MenuAction.LIST.value},
            {"name": self.strings.menu_quit, "value": MenuAction.QUIT.value},
        ]

    def display_menu(self) -> Any:
        return questionary.select(self.strings.menu_prompt, choices=self.get_menu_choices()).ask()

    def create_account(self) -> None:
        self.clean_terminal()
        print(self.strings.menu_create)
        username = input(self.strings.prompt_username)
        if username == "":
            return
        password = self.get_pass(self.strings.prompt_password)
        cookies = input(self.strings.prompt_cookies)
        self.account_manager.create(username, password, cookies, self.public_key)

    def read_account(self) -> None:
        self.clean_terminal()
        print(self.strings.menu_read)
        username = input(self.strings.prompt_username)
        account = self.account_manager.read(username)
        if account:
            ordered_dict = OrderedDict()
            ordered_dict["username"] = username
            ordered_dict["encrypted_password"] = account["encrypted_password"]
            for key, value in account.items():
                if key not in ["username", "encrypted_password"]:
                    ordered_dict[key] = value

            for key, value in ordered_dict.items():
                print(f"{key}: {value}")
        else:
            print(self.strings.msg_account_not_found)

    def update_account(self) -> None:
        self.clean_terminal()
        print(self.strings.menu_update)
        old_username = input(self.strings.prompt_old_username)
        password = self.get_pass(self.strings.prompt_password)
        if not self.account_manager.verify_password(old_username, password, self.private_key):
            return
        new_username = input(self.strings.prompt_new_username)
        password = self.get_pass(self.strings.prompt_new_password)
        cookies = input(self.strings.prompt_cookies)
        self.account_manager.edit(self.public_key, old_username, new_username, password, cookies)

    def delete_account(self) -> None:
        self.clean_terminal()
        print(self.strings.menu_delete)
        username = input(self.strings.prompt_username)
        if username in self.account_manager.accounts:
            password = self.get_pass(self.strings.prompt_password)
            if not self.account_manager.verify_password(username, password, self.private_key):
                return

            confirm_delete = questionary.select(
                self.strings.confirm_delete.format(username=username),
                choices=[
                    self.strings.confirm_yes,
                    self.strings.CONFIRM_NO,
                ],
            ).ask()

            if confirm_delete == self.strings.confirm_yes:
                self.account_manager.delete(username)
            else:
                print(self.strings.msg_operation_canceled)
        else:
            print(self.strings.msg_account_not_found)

    def password_test(self) -> None:
        self.clean_terminal()
        print(self.strings.menu_password)
        username = input(self.strings.prompt_username)
        account = self.account_manager.read(username)
        if account:
            password = self.get_pass(self.strings.prompt_password_test)
            self.account_manager.verify_password(username, password, self.private_key)
        else:
            print(self.strings.msg_account_not_found)

    def list_accounts(self) -> None:
        self.clean_terminal()
        print(self.strings.menu_list)
        accounts = self.account_manager.accounts
        if accounts:
            for username, info in accounts.items():
                print(
                    self.strings.LIST_FORMAT.format(
                        username=username,
                        quota=info["exceed_quota"],
                        time=info["exceed_time"],
                        created_at=info["created_at"],
                        cookies=info["cookies"],
                    ),
                )
        else:
            print(self.strings.msg_no_accounts)

    def execute_action(self, choice: str) -> bool:
        if choice == MenuAction.QUIT.value:
            self.clean_terminal()
            print(self.strings.msg_exit)
            return True

        action = self.action_map.get(choice)
        if action:
            action(self)
        else:
            print(self.strings.msg_invalid_choice)
        return False

    @classmethod
    def initialize_action_map(cls) -> None:
        cls.action_map = {
            MenuAction.CREATE.value: cls.create_account,
            MenuAction.READ.value: cls.read_account,
            MenuAction.UPDATE.value: cls.update_account,
            MenuAction.DELETE.value: cls.delete_account,
            MenuAction.PASSWORD.value: cls.password_test,
            MenuAction.LIST.value: cls.list_accounts,
        }

    def run(self) -> None:
        self.clean_terminal()
        while True:
            choice = self.display_menu()
            if self.execute_action(choice):
                break
        sys.exit(0)


def cli(encrypt_config: EncryptionConfig) -> None:
    cli = AccountManagerCLI(encrypt_config)
    cli.initialize_action_map()
    cli.run()



from .cookies import load_cookies
from .drission_bot import DrissionBot
from .get import get_bot
from .selenium_bot import SeleniumBot

__all__ = ["DrissionBot", "SeleniumBot", "get_bot", "load_cookies"]



import sys

if sys.version_info < (3, 10):
    raise ImportError(
        "You are using an unsupported version of Python. Only Python versions 3.10 and above are supported by v2dl",
    )

import sys
import logging
from argparse import Namespace as NamespaceT

from . import cli, common, core, utils, version, web_bot

__all__ = ["cli", "common", "core", "utils", "version", "version", "web_bot"]


def process_input(args: NamespaceT) -> common._types.BaseConfig:
    if args.version:
        print(version.__version__)  # noqa
        sys.exit(0)

    if args.input_file:
        utils.DownloadPathTool.check_input_file(args.input_file)

    base_config = common.BaseConfigManager(common.DEFAULT_CONFIG).load()

    if args.account:
        cli.cli(base_config.encryption)

    # update base_config
    if args.min_scroll is not None:
        base_config.download.min_scroll_length = args.min_scroll
    if args.max_scroll is not None:
        base_config.download.max_scroll_length = args.max_scroll

    # suppress httpx INFO level log
    level = logging.DEBUG if args.log_level == logging.DEBUG else logging.WARNING
    logging.getLogger("httpx").setLevel(level)
    logging.getLogger("httpcore").setLevel(level)
    return base_config


def create_runtime_config(
    args: NamespaceT,
    base_config: common.BaseConfig,
    logger: logging.Logger,
    log_level: int,
    service_type: utils.ServiceType = utils.ServiceType.ASYNC,
) -> common.RuntimeConfig:
    """Create runtime configuration with integrated download service and function."""

    download_service = utils.TaskServiceFactory.create(
        service_type=service_type,
        logger=logger,
        max_workers=args.concurrency,
    )

    download_api = utils.DownloadAPIFactory.create(
        service_type=service_type,
        headers=common.const.HEADERS,
        rate_limit=base_config.download.rate_limit,
        no_skip=args.no_skip,
        logger=logger,
    )

    download_function = (
        download_api.download_async
        if service_type == utils.ServiceType.ASYNC
        else download_api.download
    )
    logger.debug("using download function name: %s", download_function.__name__)

    return common.RuntimeConfig(
        url=args.url,
        input_file=args.input_file,
        bot_type=args.bot_type,
        chrome_args=args.chrome_args,
        user_agent=args.user_agent,
        use_chrome_default_profile=args.use_default_chrome_profile,
        terminate=args.terminate,
        download_service=download_service,
        download_function=download_function,
        dry_run=args.dry_run,
        logger=logger,
        log_level=log_level,
        no_skip=args.no_skip,
    )


def main() -> int:
    args = cli.parse_arguments()
    base_config = process_input(args)

    logger = common.setup_logging(
        args.log_level,
        log_path=base_config.paths.system_log,
        logger_name=version.__package_name__,
    )
    runtime_config = create_runtime_config(args, base_config, logger, args.log_level)

    web_bot_ = web_bot.get_bot(runtime_config, base_config)
    scraper = core.ScrapeManager(runtime_config, base_config, web_bot_)
    scraper.start_scraping()

    return 0



import os
import re
import sys
import time
import asyncio
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from mimetypes import guess_extension
from pathlib import Path

import httpx
from pathvalidate import sanitize_filename
from requests import Response

from .parser import LinkParser
from ..common._types import PathType

logger = logging.getLogger()


class BaseDownloadAPI(ABC):
    """Base protocol for download APIs."""

    def __init__(
        self,
        headers: dict[str, str],
        rate_limit: int,
        no_skip: bool,
        logger: logging.Logger,
    ):
        self.headers = headers
        self.rate_limit = rate_limit
        self.no_skip = no_skip
        self.logger = logger

    @abstractmethod
    def download(self, album_name: str, url: str, filename: str, base_folder: Path) -> bool:
        """Synchronous download method."""
        raise NotImplementedError

    @abstractmethod
    async def download_async(
        self,
        task_id: str,
        url: str,
        filename: str,
        destination: Path,
    ) -> bool:
        """Asynchronous download method."""
        raise NotImplementedError


class ImageDownloadAPI(BaseDownloadAPI):
    """Image download implementation."""

    def download(self, album_name: str, url: str, filename: str, base_folder: Path) -> bool:
        try:
            album_name = album_name.rsplit("_", 1)[0]
            file_path = DownloadPathTool.get_file_dest(base_folder, album_name, filename)
            DownloadPathTool.mkdir(file_path.parent)

            if DownloadPathTool.is_file_exists(file_path, self.no_skip, self.logger):
                return True

            Downloader.download(url, file_path, self.headers, self.rate_limit)
            self.logger.info("Downloaded: '%s'", file_path)
            return True
        except Exception as e:
            self.logger.error("Error in threaded task '%s': %s", url, e)
            return False

    async def download_async(
        self,
        album_name: str,
        url: str,
        filename: str,
        base_folder: Path,
    ) -> bool:
        try:
            album_name = album_name.rsplit("_", 1)[0]
            file_path = DownloadPathTool.get_file_dest(base_folder, album_name, filename)
            DownloadPathTool.mkdir(file_path.parent)

            if DownloadPathTool.is_file_exists(file_path, self.no_skip, self.logger):
                return True

            await Downloader.download_async(url, file_path, self.headers, self.rate_limit)
            self.logger.info("Downloaded: '%s'", file_path)
            return True
        except Exception as e:
            self.logger.error("Error in async task '%s': %s", album_name, e)
            return False


class VideoDownloadAPI(BaseDownloadAPI):
    """Video download implementation."""

    def download(self, task_id: str, url: str, resp: Response, destination: Path) -> bool:
        raise NotImplementedError

    async def download_async(
        self,
        task_id: str,
        url: str,
        resp: Response,
        destination: Path,
    ) -> bool:
        raise NotImplementedError


class ActorDownloadAPI(BaseDownloadAPI):
    """Actor-based download implementation."""

    def download(self, album_name: str, url: str, alt: str, base_folder: Path) -> bool:
        raise NotImplementedError

    async def download_async(self, task_id: str, url: str, alt: str, destination: Path) -> bool:
        raise NotImplementedError


class Downloader:
    """Handles file downloading operations."""

    @staticmethod
    def download(
        url: str,
        save_path: Path,
        headers: dict[str, str] | None,
        speed_limit_kbps: int,
    ) -> None:
        """Download with speed limit."""
        if headers is None:
            headers = {}
        chunk_size = 1024
        speed_limit_bps = speed_limit_kbps * 1024

        timeout = httpx.Timeout(10.0, read=5.0)
        with httpx.Client(timeout=timeout) as client:
            with client.stream("GET", url, headers=headers) as response:
                response.raise_for_status()
                ext = "." + DownloadPathTool.get_ext(response)
                save_path = save_path.with_suffix(ext)

                with open(save_path, "wb") as file:
                    start_time = time.time()
                    downloaded = 0
                    for chunk in response.iter_bytes(chunk_size=chunk_size):
                        file.write(chunk)
                        downloaded += len(chunk)
                        elapsed_time = time.time() - start_time
                        expected_time = downloaded / speed_limit_bps
                        if elapsed_time < expected_time:
                            time.sleep(expected_time - elapsed_time)

    @staticmethod
    async def download_async(
        url: str,
        save_path: Path,
        headers: dict[str, str] | None,
        speed_limit_kbps: int,
    ) -> None:
        """Asynchronous download with speed limit."""
        if headers is None:
            headers = {}
        chunk_size = 1024
        speed_limit_bps = speed_limit_kbps * 1024

        timeout = httpx.Timeout(10.0, read=30.0)
        async with httpx.AsyncClient(timeout=timeout) as client:
            async with client.stream("GET", url, headers=headers) as response:
                response.raise_for_status()
                ext = "." + DownloadPathTool.get_ext(response)
                save_path = save_path.with_suffix(ext)

                with open(save_path, "wb") as file:
                    start_time = asyncio.get_event_loop().time()
                    downloaded = 0
                    async for chunk in response.aiter_bytes(chunk_size=chunk_size):
                        file.write(chunk)
                        downloaded += len(chunk)
                        elapsed_time = asyncio.get_event_loop().time() - start_time
                        expected_time = downloaded / speed_limit_bps
                        if elapsed_time < expected_time:
                            await asyncio.sleep(expected_time - elapsed_time)


class DownloadPathTool:
    """Handles file and directory operations."""

    @staticmethod
    def mkdir(folder_path: PathType) -> None:
        """Ensure the folder exists, create it if not."""
        Path(folder_path).mkdir(parents=True, exist_ok=True)

    @staticmethod
    def is_file_exists(file_path: PathType, no_skip: bool, logger: logging.Logger) -> bool:
        """Check if the file exists and log the status."""
        if Path(file_path).exists() and not no_skip:
            logger.info("File already exists: '%s'", file_path)
            return True
        return False

    @staticmethod
    def get_file_dest(
        download_root: PathType,
        album_name: str,
        filename: str,
        extension: str | None = None,
    ) -> Path:
        """Construct the file path for saving the downloaded file.

        Args:
            download_root (PathType): The base download folder for v2dl
            album_name (str): The name of the download album, used for the sub-directory
            filename (str): The name of the target download file
            extension (str | None): The file extension of the target download file
        Returns:
            PathType: The full path of the file
        """
        ext = f".{extension}" if extension else ""
        folder = Path(download_root) / sanitize_filename(album_name)
        sf = sanitize_filename(filename)
        return folder / f"{sf}{ext}"

    @staticmethod
    def get_image_ext(url: str, default_ext: str = "jpg") -> str:
        """Get the extension of a URL."""
        image_extensions = r"\.(jpg|jpeg|png|gif|bmp|webp|tiff|svg)(?:\?.*|#.*|$)"
        match = re.search(image_extensions, url, re.IGNORECASE)
        if match:
            # Normalize 'jpeg' to 'jpg'
            return "jpg" if match.group(1).lower() == "jpeg" else match.group(1).lower()
        return default_ext

    @staticmethod
    def get_ext(
        response: httpx.Response,
        default_method: Callable[[str, str], str] | None = None,
    ) -> str:
        """Guess file extension based on response Content-Type."""
        if default_method is None:
            default_method = DownloadPathTool.get_image_ext

        content_type = response.headers.get("Content-Type", "").split(";")[0].strip()
        extension = guess_extension(content_type)
        if extension:
            return extension.lstrip(".")

        return default_method(str(response.url), "jpg")

    @staticmethod
    def check_input_file(input_path: PathType) -> None:
        if input_path and not os.path.isfile(input_path):
            logging.error("Input file %s does not exist.", input_path)
            sys.exit(1)
        else:
            logging.info("Input file %s exists and is accessible.", input_path)


class AlbumTracker:
    """Download log in units of albums."""

    def __init__(self, download_log: str):
        self.album_log_path = download_log

    def is_downloaded(self, album_url: str) -> bool:
        if os.path.exists(self.album_log_path):
            with open(self.album_log_path) as f:
                downloaded_albums = f.read().splitlines()
            return album_url in downloaded_albums
        return False

    def log_downloaded(self, album_url: str) -> None:
        album_url = LinkParser.remove_page_num(album_url)
        if not self.is_downloaded(album_url):
            with open(self.album_log_path, "a") as f:
                f.write(album_url + "\n")


def download_album(
    album_name: str,
    file_links: list[tuple[str, str]],
    destination: str,
    headers: dict[str, str],
    rate_limit: int,
    no_skip: bool,
    logger: logging.Logger,
) -> None:
    """Basic usage example: download files from a list of links."""
    task_manager = ImageDownloadAPI(
        headers=headers,
        rate_limit=rate_limit,
        no_skip=no_skip,
        logger=logger,
    )
    for url, alt in file_links:
        task_id = f"{album_name}_{alt}"
        task_manager.download(task_id, url, alt, Path(destination))



# v2dl/common/__init__.py
from ._types import (
    BaseConfig,
    ChromeConfig,
    DownloadConfig,
    EncryptionConfig,
    PathConfig,
    RuntimeConfig,
)
from .config import BaseConfigManager
from .const import DEFAULT_CONFIG, SELENIUM_AGENT
from .error import BotError, DownloadError, FileProcessingError, ScrapeError, SecurityError
from .logger import setup_logging

__all__ = [
    "DEFAULT_CONFIG",
    "SELENIUM_AGENT",
    "BaseConfig",
    "BaseConfigManager",
    "BotError",
    "ChromeConfig",
    "DownloadConfig",
    "DownloadError",
    "EncryptionConfig",
    "FileProcessingError",
    "PathConfig",
    "RuntimeConfig",
    "ScrapeError",
    "SecurityError",
    "setup_logging",
]



import re
from logging import Logger
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from lxml import html


class LinkParser:
    """Tool class for parsing and modifying URLs."""

    @staticmethod
    def parse_input_url(url: str) -> tuple[list[str], int]:
        """
        Extracts path segments and the starting page number from a URL.

        Args:
            url (str): Input URL.

        Returns:
            tuple[list[str], int]: Path segments and the starting page number.
        """
        parsed_url = urlparse(url)
        path_parts = parsed_url.path.split("/")
        query_params = parse_qs(parsed_url.query)
        start_page: int = int(query_params.get("page", [1])[0])  # default page=1
        return path_parts, start_page

    @staticmethod
    def parse_html(html_content: str, logger: Logger) -> html.HtmlElement | None:
        """Parses HTML content into an HTML element.

        Args:
            html_content (str): HTML content as a string.
            logger (Logger): Logger for error handling.

        Returns:
            html.HtmlElement | None: Parsed HTML element or None if parsing fails.
        """
        if "Failed" in html_content:
            return None

        try:
            return html.fromstring(html_content)
        except Exception as e:
            logger.error("Error parsing HTML content: %s", e)
            return None

    @staticmethod
    def get_max_page(tree: html.HtmlElement) -> int:
        """
        Retrieves the maximum page number from a pagination element.

        Args:
            tree (html.HtmlElement): Parsed HTML tree.

        Returns:
            int: Maximum page number, default is 1 if none found.
        """
        page_links = tree.xpath(
            '//li[@class="page-item"]/a[@class="page-link" and string-length(text()) <= 2]/@href',
        )

        if not page_links:
            return 1

        page_numbers = []
        for link in page_links:
            match = re.search(r"page=(\d+)", link)
            if match:
                page_number = int(match.group(1))
            else:
                page_number = 1
            page_numbers.append(page_number)

        return max(page_numbers)

    @staticmethod
    def add_page_num(url: str, page: int) -> str:
        """
        Adds or updates the page number in a URL.

        Args:
            url (str): Original URL.
            page (int): Page number to add or update.

        Returns:
            str: Updated URL with the specified page number.
        """
        parsed_url = urlparse(url)  # 解析 URL
        query_params = parse_qs(parsed_url.query)  # 解析查詢參數
        query_params["page"] = [str(page)]  # 修改頁碼

        new_query = urlencode(query_params, doseq=True)  # 組合成字串
        new_url = parsed_url._replace(query=new_query)  # 替換頁碼

        # Example
        # url = "https://example.com/search?q=test&sort=asc", page = 3
        # parsed_url: ParseResult(scheme='https', netloc='example.com', path='/search', params='', query='q=test&sort=asc', fragment='')
        # query_params: {'q': ['test'], 'sort': ['asc'], 'page': ['3']}
        # new_query: 'q=test&sort=asc&page=3'
        # new_url: ParseResult(scheme='https', netloc='example.com', path='/search', params='', query='q=test&sort=asc&page=3', fragment='')
        # urlunparse: 'https://example.com/search?q=test&sort=asc&page=3'
        return urlunparse(new_url)

    @staticmethod
    def remove_page_num(url: str) -> str:
        """
        Removes the page parameter from a URL.

        Args:
            url (str): Original URL.

        Returns:
            str: URL without the page parameter.
        """
        parsed_url = urlparse(url)
        query_params = parse_qs(parsed_url.query)

        if "page" in query_params:
            del query_params["page"]

        new_query = urlencode(query_params, doseq=True)

        return urlunparse(parsed_url._replace(query=new_query))



import logging
import argparse


def parse_arguments() -> argparse.Namespace:
    formatter = lambda prog: argparse.HelpFormatter(prog, max_help_position=36)
    parser = argparse.ArgumentParser(description="V2PH scraper.", formatter_class=formatter)

    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument("url", nargs="?", help="URL to scrape")
    input_group.add_argument(
        "-i",
        "--input-file",
        metavar="PATH",
        help="Path to txt file containing URL list to be downloaded",
    )
    input_group.add_argument("-a", "--account", action="store_true", help="Manage account")
    input_group.add_argument("--version", action="store_true", help="Show package version")

    parser.add_argument(
        "--bot",
        dest="bot_type",
        default="drission",
        type=str,
        choices=["selenium", "drission"],
        required=False,
        help="Type of bot to use (default: drission)",
    )

    parser.add_argument("--concurrency", default=5, type=int, help="maximum download concurrency")
    parser.add_argument("--min-scroll", type=int, help="minimum scroll length of web bot")
    parser.add_argument("--max-scroll", type=int, help="maximum scroll length of web bot")

    parser.add_argument(
        "--chrome-args",
        type=str,
        help="Override Chrome arguments (example: --chrome-args='--arg1//--arg2//--arg3')",
    )
    parser.add_argument(
        "--user-agent",
        type=str,
        help="Override user-agent (example: --user-agent='Mozilla/5.0 (Windows NT 10.0; Win64; x64)...')",
    )

    parser.add_argument("--dry-run", action="store_true", help="Dry run without downloading")
    parser.add_argument("--no-skip", action="store_true", help="Do not skip downloaded files")
    parser.add_argument("--terminate", action="store_true", help="Terminate chrome after scraping")
    parser.add_argument(
        "--use-default-chrome-profile",
        action="store_true",
        help="Use default chrome profile. Using default profile with an operating chrome is not valid",
    )

    log_group = parser.add_mutually_exclusive_group()
    log_group.add_argument("-q", "--quiet", action="store_true", help="Quiet mode")
    log_group.add_argument("-v", "--verbose", action="store_true", help="Verbose mode")
    log_group.add_argument(
        "--log-level",
        default=None,
        type=int,
        choices=range(1, 6),
        help="Set log level (1~5)",
    )

    args = parser.parse_args()

    if args.quiet:
        log_level = logging.ERROR
    elif args.verbose:
        log_level = logging.DEBUG
    elif args.log_level is not None:
        log_level_mapping = {
            1: logging.DEBUG,
            2: logging.INFO,
            3: logging.WARNING,
            4: logging.WARNING,
            5: logging.CRITICAL,
        }
        log_level = log_level_mapping.get(args.log_level, logging.INFO)
    else:
        log_level = logging.INFO

    args.chrome_args = args.chrome_args.split("//") if args.chrome_args else None
    args.log_level = log_level
    return args



from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from logging import Logger
from typing import Any

from .download import (
    ActorDownloadAPI,
    BaseDownloadAPI,
    ImageDownloadAPI,
    VideoDownloadAPI,
)
from .multitask import AsyncService, BaseTaskService, ThreadingService


@dataclass
class Task:
    """Unified task container for both threading and async services."""

    task_id: str
    func: Callable[..., Any]
    args: tuple[Any, ...] = ()
    kwargs: dict[str, Any] | None = None

    def __post_init__(self) -> None:
        self.kwargs = self.kwargs or {}


class ServiceType(Enum):
    """Service type enumeration."""

    THREADING = "threading"
    ASYNC = "async"
    ACTOR = "actor"


class MediaType(Enum):
    """Media type enumeration."""

    IMAGE = "image"
    VIDEO = "video"


class TaskServiceFactory:
    """Factory class for creating task services."""

    @staticmethod
    def create(
        service_type: ServiceType,
        logger: Logger,
        max_workers: int = 5,
    ) -> BaseTaskService:
        """Create a new task service instance."""
        if service_type == ServiceType.THREADING:
            return ThreadingService(logger, max_workers)
        elif service_type == ServiceType.ASYNC:
            return AsyncService(logger, max_workers)
        else:
            raise ValueError(f"Unknown service type: {service_type}")


class DownloadAPIFactory:
    """Factory for creating download API instances."""

    _api_registry: dict[ServiceType, type[BaseDownloadAPI]] = {
        ServiceType.THREADING: ImageDownloadAPI,
        ServiceType.ASYNC: ImageDownloadAPI,
        ServiceType.ACTOR: ActorDownloadAPI,
    }

    @classmethod
    def create(
        cls,
        service_type: ServiceType,
        headers: dict[str, str],
        rate_limit: int,
        no_skip: bool,
        logger: Logger,
        media_type: MediaType = MediaType.IMAGE,
    ) -> BaseDownloadAPI:
        """Create a download API instance based on service type and media type."""
        api_class = cls._api_registry.get(service_type)
        if not api_class:
            raise ValueError(f"Unknown service type: {service_type}")

        if media_type == MediaType.VIDEO:
            return VideoDownloadAPI(headers, rate_limit, no_skip, logger)

        return api_class(headers, rate_limit, no_skip, logger)



import os
import logging
from typing import ClassVar

from colorama import Fore, Style, init


class CustomFormatter(logging.Formatter):
    COLORS: ClassVar[dict[int, str]] = {
        logging.DEBUG: Fore.LIGHTBLACK_EX,
        logging.INFO: Fore.WHITE,
        logging.WARNING: Fore.YELLOW,
        logging.ERROR: Fore.RED,
        logging.CRITICAL: Fore.RED + Style.BRIGHT,
    }
    GREEN = Fore.GREEN
    RESET = Style.RESET_ALL

    def __init__(self, use_color: bool = True) -> None:
        super().__init__()
        init()
        self.use_color = use_color

    def format(self, record: logging.LogRecord) -> str:
        if self.use_color:
            color = self.COLORS.get(record.levelno, self.RESET)
            levelname = record.levelname.lower()
            return f"[{self.GREEN}{self.formatTime(record, '%H:%M:%S')}{self.RESET}][{color}{levelname}{self.RESET}] - {record.getMessage()}"
        else:
            # Convert levelname to lowercase for file logs
            levelname = record.levelname.lower()
            return f"[{self.formatTime(record, '%H:%M:%S')}][{levelname}] - {record.getMessage()}"


def setup_logging(
    level: int,
    log_path: str | None = None,
    logger_name: str | None = None,
    archive: bool = True,
) -> logging.Logger:
    """Configure logging with console and optional file handlers.

    Args:
        level (int): Logging level (e.g., logging.DEBUG).
        log_path (str): Path to the log file.
        archive (bool): If True, enables file logging.
    """
    logger = logging.getLogger(logger_name)

    # Clear existing handlers
    logging.root.handlers.clear()
    logger.handlers.clear()

    # Create formatters
    color_formatter = CustomFormatter(use_color=True)
    if archive:
        plain_formatter = CustomFormatter(use_color=False)

    # Console handler
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(color_formatter)
    logging.root.addHandler(console_handler)

    # File handler
    if archive and log_path:
        log_dir = os.path.dirname(log_path)
        if log_dir:
            os.makedirs(log_dir, exist_ok=True)
        file_handler = logging.FileHandler(log_path, encoding="utf-8")
        file_handler.setFormatter(plain_formatter)
        logging.root.addHandler(file_handler)

    # Set log level
    logging.root.setLevel(level)
    logger.setLevel(level)

    return logger


if __name__ == "__main__":
    # Set up logging
    setup_logging(logging.DEBUG, "test_logger.log")

    # Create a logger
    logger = logging.getLogger(__name__)

    # Log messages to test the configuration
    logger.debug("Debug message")
    logger.info("This is an info message.")
    logger.warning("Warning message")
    logger.error("This is an error message.")
    logger.critical("Critical message")



import os
import sys
import atexit
import base64
import ctypes
import random
import secrets
import threading
from dataclasses import dataclass
from datetime import datetime, timedelta
from logging import Logger
from typing import Any, Literal, overload

import yaml
from dotenv import load_dotenv, set_key
from nacl.public import PrivateKey, PublicKey, SealedBox
from nacl.pwhash import argon2id
from nacl.secret import SecretBox
from nacl.utils import EncryptedMessage, random as nacl_random

from ..common import BaseConfigManager, EncryptionConfig, SecurityError


@dataclass
class KeyPair:
    private_key: PrivateKey
    public_key: PublicKey


class Encryptor:
    """Managing encryption and decryption operations."""

    def __init__(self, logger: Logger, encrypt_config: EncryptionConfig) -> None:
        self.logger = logger
        self.encrypt_config = encrypt_config

    def encrypt_master_key(self, master_key: bytes) -> tuple[bytes, bytes, bytes]:
        salt = secrets.token_bytes(self.encrypt_config.salt_bytes)
        encryption_key = secrets.token_bytes(self.encrypt_config.key_bytes)
        derived_key = self.derive_key(encryption_key, salt)

        box = SecretBox(derived_key)
        nonce = nacl_random(self.encrypt_config.nonce_bytes)
        encrypted_master_key = box.encrypt(master_key, nonce)

        derived_key = bytearray(len(derived_key))
        self.logger.info("Master key encryption successful")
        return encrypted_master_key, salt, encryption_key

    def decrypt_master_key(
        self,
        encrypted_master_key: bytes,
        salt: str,
        encryption_key: str,
    ) -> bytes:
        salt_b64 = base64.b64decode(salt)
        enc_key_b64 = base64.b64decode(encryption_key)
        derived_key = self.derive_key(enc_key_b64, salt_b64)
        box = SecretBox(derived_key)

        master_key = box.decrypt(encrypted_master_key)

        self.logger.info("Master key decryption successful")
        return master_key

    def encrypt_private_key(self, private_key: PrivateKey, master_key: bytes) -> EncryptedMessage:
        box = SecretBox(master_key)
        nonce = nacl_random(self.encrypt_config.nonce_bytes)
        return box.encrypt(private_key.encode(), nonce)

    def decrypt_private_key(self, encrypted_private_key: bytes, master_key: bytes) -> PrivateKey:
        box = SecretBox(master_key)
        private_key_bytes = box.decrypt(encrypted_private_key)
        private_key = PrivateKey(private_key_bytes)
        cleanup([private_key_bytes])
        return private_key

    def encrypt_password(self, password: str, public_key: PublicKey) -> str:
        sealed_box = SealedBox(public_key)
        encrypted = sealed_box.encrypt(password.encode())
        self.logger.info("Password encryption successful")
        return base64.b64encode(encrypted).decode("utf-8")

    def decrypt_password(self, encrypted_password: str, private_key: PrivateKey) -> str:
        try:
            encrypted = base64.b64decode(encrypted_password)
            sealed_box = SealedBox(private_key)
            decrypted = sealed_box.decrypt(encrypted)
            return decrypted.decode()
        except Exception as e:
            self.logger.error("Password decryption failed: %s", str(e))
            raise SecurityError from e

    def derive_key(self, encryption_key: bytes, salt: bytes) -> bytes:
        return argon2id.kdf(
            self.encrypt_config.key_bytes,
            encryption_key,
            salt,
            opslimit=self.encrypt_config.kdf_ops_limit,
            memlimit=self.encrypt_config.kdf_mem_limit,
        )

    def validate_keypair(self, private_key: PrivateKey, public_key: PublicKey) -> None:
        try:
            test_data = b"test"
            sealed_box = SealedBox(public_key)
            sealed_box_priv = SealedBox(private_key)

            encrypted = sealed_box.encrypt(test_data)
            decrypted = sealed_box_priv.decrypt(encrypted)

            if decrypted != test_data:
                raise SecurityError
        except Exception as e:
            self.logger.error("Key pair validation failed: %s", str(e))
            raise SecurityError from e


class KeyIOHelper(Encryptor):
    """Manage the loading, saving, and validation of cryptographic keys."""

    def __init__(
        self,
        logger: Logger,
        path_config: dict[str, str] | None,
        encrypt_config: EncryptionConfig,
    ) -> None:
        super().__init__(logger, encrypt_config)
        self.logger = logger
        self.path_config = self.init_conf(path_config)

    def init_conf(self, path_config: dict[str, str] | None) -> dict[str, str]:
        if path_config is None:
            base_dir = BaseConfigManager.get_system_config_dir()
            self.logger.debug("Initializing config with base directory: %s", base_dir)
            return {
                "key_folder": os.path.join(base_dir, ".keys"),
                "env_path": os.path.join(base_dir, ".env"),
                "master_key_file": os.path.join(base_dir, ".keys", "master_key.enc"),
                "private_key_file": os.path.join(base_dir, ".keys", "private_key.pem"),
                "public_key_file": os.path.join(base_dir, ".keys", "public_key.pem"),
            }
        else:
            return path_config

    def load_keys(self) -> KeyPair:
        self.logger.debug("Loading and validating keys")
        master_key = self.load_master_key()
        private_key = self.load_private_key(master_key)
        public_key = self.load_public_key()

        self.validate_keypair(private_key, public_key)
        cleanup([master_key])

        self.logger.info("Keys loaded and validated successfully")
        return KeyPair(private_key, public_key)

    def load_secret(self, env_path: str) -> tuple[str, str]:
        """Load and validate salt and encryption_key from .env file."""
        load_dotenv(env_path)
        salt_base64 = SecureFileHandler.read_env("SALT")
        encryption_key_base64 = SecureFileHandler.read_env("ENCRYPTION_KEY")
        return salt_base64, encryption_key_base64

    def load_master_key(self, path: str | None = None) -> bytes:
        _path = self.path_config["master_key_file"] if path is None else path
        encrypted_master_key = SecureFileHandler.read_file(_path, False)
        salt, encryption_key = self.load_secret(self.path_config["env_path"])
        return self.decrypt_master_key(encrypted_master_key, salt, encryption_key)

    def load_public_key(self, path: str | None = None) -> PublicKey:
        _path = self.path_config["public_key_file"] if path is None else path
        public_key_bytes = SecureFileHandler.read_file(_path, False)
        return PublicKey(public_key_bytes)

    def load_private_key(self, master_key: bytes, path: str | None = None) -> PrivateKey:
        _path = self.path_config["private_key_file"] if path is None else path
        encrypted_private_key = SecureFileHandler.read_file(_path, False)
        return self.decrypt_private_key(encrypted_private_key, master_key)

    def save_keys(self, keys: tuple[bytes, bytes, PublicKey, bytes, bytes]) -> None:
        SecureFileHandler.write_file(self.path_config["master_key_file"], keys[0])
        SecureFileHandler.write_file(self.path_config["private_key_file"], keys[1])
        SecureFileHandler.write_file(self.path_config["public_key_file"], keys[2].encode(), 0o644)
        SecureFileHandler.write_env(self.path_config["env_path"], "SALT", keys[3])
        SecureFileHandler.write_env(self.path_config["env_path"], "ENCRYPTION_KEY", keys[4])

    def check_folder(self) -> None:
        if not os.path.exists(self.path_config["key_folder"]):
            os.makedirs(self.path_config["key_folder"], mode=0o700)
            self.logger.info("Secure folder created at %s", self.path_config["key_folder"])
        elif not self.check_permission(self.path_config["key_folder"]):
            os.chmod(self.path_config["key_folder"], 0o700)
            self.logger.info("Permissions updated for folder at %s", self.path_config["key_folder"])

    def check_permission(self, folder_path: str) -> bool:
        folder_permission = 0o700
        current_permissions = os.stat(folder_path).st_mode & 0o777
        return current_permissions == folder_permission


class KeyManager(KeyIOHelper):
    """Top level class managing key generation."""

    def __init__(
        self,
        logger: Logger,
        encrypt_config: EncryptionConfig,
        path_dict: dict[str, str] | None = None,
    ) -> None:
        super().__init__(logger, path_dict, encrypt_config)
        self.check_folder()

        keys = self._init_keys()
        if keys is not None:
            self.save_keys(keys)

    def _init_keys(self) -> tuple[bytes, bytes, PublicKey, bytes, bytes] | None:
        if self._keys_exist():
            self.logger.info("Key pair already exists")
            return None

        self.check_folder()
        return self._generate_and_encrypt_keys()

    def _keys_exist(self) -> bool:
        return os.path.exists(self.path_config["private_key_file"]) and os.path.exists(
            self.path_config["public_key_file"],
        )

    def _generate_and_encrypt_keys(self) -> tuple[bytes, bytes, PublicKey, bytes, bytes]:
        keys = self._generate_key_pair()
        master_key = secrets.token_bytes(self.encrypt_config.key_bytes)
        encrypted_master_key, salt, encryption_key = self.encrypt_master_key(master_key)
        encrypted_private_key = self.encrypt_private_key(keys.private_key, master_key)

        cleanup([master_key])
        self.logger.info("Key pair has been successfully generated")
        return (
            encrypted_master_key,
            encrypted_private_key,
            keys.public_key,
            salt,
            encryption_key,
        )

    def _generate_key_pair(self) -> KeyPair:
        private_key = PrivateKey.generate()
        return KeyPair(private_key, private_key.public_key)


class AccountManager:
    MAX_QUOTA = 16

    def __init__(self, logger: Logger, key_manager: KeyManager, yaml_path: str = ""):
        self.logger = logger
        self.yaml_path = (
            yaml_path
            if yaml_path
            else os.path.join(BaseConfigManager.get_system_config_dir(), "accounts.yaml")
        )
        self.key_manager = key_manager
        self.lock = threading.RLock()
        self.accounts, self.runtime_state = self._load_yaml()
        self.check()
        atexit.register(self._save_yaml)

    def create(self, username: str, password: str, cookies: str, public_key: PublicKey) -> None:
        with self.lock:
            encrypted_password = self.key_manager.encrypt_password(password, public_key)
            self.accounts[username] = {
                "encrypted_password": encrypted_password,
                "created_at": datetime.now().strftime("%Y-%m-%dT%H:%M:%S"),
                "exceed_quota": False,
                "exceed_time": "",
                "cookies": cookies,
            }
        self.logger.info("Account %s has been created.", username)
        self._save_yaml()

    def delete(self, username: str) -> None:
        with self.lock:
            if username in self.accounts:
                del self.accounts[username]
                self.logger.info("Account %s has been deleted.", username)
            else:
                self.logger.error("Account %s not found.", username)
            self._save_yaml()

    def read(self, username: str) -> dict[str, Any] | None:
        return self.accounts.get(username)

    def edit(
        self,
        public_key: PublicKey,
        old_username: str,
        new_username: str | None,
        new_password: str | None,
        new_cookies: str | None,
    ) -> None:
        with self.lock:
            if old_username in self.accounts:
                if new_username:
                    self.accounts[new_username] = self.accounts.pop(old_username)
                if new_password:
                    encrypted_password = self.key_manager.encrypt_password(new_password, public_key)
                    self.accounts[new_username or old_username]["encrypted_password"] = (
                        encrypted_password
                    )
                if new_cookies:
                    self.accounts[new_username or old_username]["cookies"] = new_cookies
                self.logger.info("Account %s has been updated.", old_username)
            else:
                self.logger.error("Account not found.")

    def update_account(self, username: str, field: str, new_value: Any) -> None:
        with self.lock:
            account = self.accounts.get(username)
            if account:
                if field in account:
                    account[field] = new_value
                    self.logger.info("Updated %s for account %s.", field, username)
                    self._save_yaml()
                else:
                    self.logger.error("Field '%s' does not exist in account '%s'.", field, account)
            else:
                self.logger.error("Account %s not found.", username)

    def update_runtime_state(self, username: str, field: str, value: Any) -> None:
        account = self.runtime_state.get(username)
        if account:
            if field in account:
                account[field] = value
                self.logger.info("Updated %s for account %s.", field, username)
            else:
                self.logger.error("Field '%s' does not exist in the account.", field)
        else:
            self.logger.error("Account %s not found.", username)

    def verify_password(self, username: str, password: str, private_key: PrivateKey) -> bool:
        account = self.accounts.get(username)
        if not account:
            self.logger.error("Account does not exist.")
            return False

        encrypted_password = account.get("encrypted_password")
        decrypted_password = self.key_manager.decrypt_password(encrypted_password, private_key)
        if decrypted_password == password:
            print("*----------------*")
            print("|Password correct|")
            print("*----------------*")
            return True
        else:
            print("*------------------*")
            print("|Incorrect password|")
            print("*------------------*")
            return False

    def check(self) -> None:
        """檢查所有帳號的 exceed_time 是否超過 24 小時，若超過則清除 exceed_time 並將重置 exceed_quota."""
        now = datetime.now()
        update = False

        for _, account in self.accounts.items():
            exceed_time = account.get("exceed_time", "")
            if exceed_time:
                exceed_time_time = datetime.strptime(exceed_time, "%Y-%m-%dT%H:%M:%S")
                if now - exceed_time_time > timedelta(hours=24):
                    account["exceed_time"] = ""
                    account["exceed_quota"] = False
                    update = True

        if update:
            self._save_yaml()

    def random_pick(self, private_key: PrivateKey) -> tuple[str, str]:
        eligible_accounts = {k: v for k, v in self.accounts.items() if not v["exceed_quota"]}

        if not eligible_accounts:
            self.logger.info("All accounts have exhausted their reading quota. Existing.")
            sys.exit(1)

        # filter invalid account during runtime based on cookies and password
        eligible_accounts = {
            k: v
            for k, v in eligible_accounts.items()
            # Allow any types of login methods
            if self.runtime_state[k].get("cookies_valid", True)
            or self.runtime_state[k].get("password_valid", True)
        }

        if not eligible_accounts:
            self.logger.info("No eligible accounts available for login. Existing.")
            sys.exit(1)

        username, account = random.choice(list(eligible_accounts.items()))
        enc_pw = account["encrypted_password"]
        dec_pw = self.key_manager.decrypt_password(enc_pw, private_key)

        return username, dec_pw

    def _save_yaml(self) -> None:
        with self.lock:
            with open(self.yaml_path, "w") as file:
                yaml.dump(self.accounts, file, default_flow_style=False)
        # self.logger.info("Successfully update accounts information.")

    def _load_yaml(self) -> tuple[dict[str, Any], dict[str, Any]]:
        try:
            with open(self.yaml_path) as file:
                account = yaml.safe_load(file) or {}
            runtime_state = self._login_state(account)
            return account, runtime_state
        except FileNotFoundError:
            return {}, {}

    def _login_state(self, accounts: dict[str, Any]) -> dict[str, Any]:
        """Store login status. Use an individual variable for not storing back to file"""
        return {
            email: {
                "cookies_valid": True,
                "password_valid": True,
            }
            for email in accounts  # email is the key of each account
        }


class SecureFileHandler:
    @staticmethod
    def write_file(path: str, data: str | bytes, permissions: int = 0o400) -> None:
        if isinstance(data, str):
            data = data.encode("utf-8")

        with open(path, "wb") as f:
            f.write(data)
        os.chmod(path, permissions)

    @staticmethod
    @overload
    def read_file(path: str, decode: Literal[True]) -> str: ...

    @staticmethod
    @overload
    def read_file(path: str, decode: Literal[False]) -> bytes: ...

    @staticmethod
    def read_file(path: str, decode: bool = False) -> str | bytes:
        with open(path, "rb") as f:
            _data = f.read()

        return _data.decode("utf-8") if decode else _data

    @staticmethod
    def write_env(env_path: str, key: str, value: str | bytes) -> None:
        if isinstance(value, bytes):
            value = base64.b64encode(value).decode("utf-8")

        load_dotenv(env_path)
        set_key(env_path, key, value)

    @staticmethod
    def read_env(key: str) -> str:
        value = os.getenv(key)
        if value is None:
            raise SecurityError(f"Missing required environment variable: {key}")
        return value


def cleanup(sensitive_data: list[bytes]) -> None:
    for data in sensitive_data:
        length = len(data)
        buffer = ctypes.create_string_buffer(length)
        ctypes.memmove(ctypes.addressof(buffer), data, length)
        ctypes.memset(ctypes.addressof(buffer), 0, length)
        del buffer



from collections.abc import Callable
from dataclasses import dataclass
from logging import Logger
from pathlib import Path
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ..utils import BaseTaskService

PathType = str | Path


@dataclass
class RuntimeConfig:
    url: str
    input_file: str
    bot_type: str
    chrome_args: list[str] | None
    user_agent: str | None
    terminate: bool
    download_service: "BaseTaskService"
    download_function: Callable[..., Any]
    dry_run: bool
    logger: Logger
    log_level: int
    no_skip: bool = False
    use_chrome_default_profile: bool = False


@dataclass
class EncryptionConfig:
    key_bytes: int
    salt_bytes: int
    nonce_bytes: int
    kdf_ops_limit: int
    kdf_mem_limit: int


@dataclass
class DownloadConfig:
    min_scroll_length: int
    max_scroll_length: int
    min_scroll_step: int
    max_scroll_step: int
    rate_limit: int
    download_dir: str


@dataclass
class PathConfig:
    download_log: str
    system_log: str


@dataclass
class ChromeConfig:
    exec_path: str
    profile_path: str


@dataclass
class BaseConfig:
    download: DownloadConfig
    paths: PathConfig
    chrome: ChromeConfig
    encryption: EncryptionConfig



import json
import logging
from http.cookiejar import MozillaCookieJar
from pathlib import Path
from typing import Any

logger = logging.getLogger()


def load_cookies(file_path: str) -> dict[str, str] | dict[str, Any]:
    if not file_path:
        logger.error("File path not exists.")
        return {}

    fp = Path(file_path)
    if not fp.exists():
        logger.error("File not found: %s", file_path)
        return {}
    if not fp.is_file():
        logger.error("Not a valid file: %s", file_path)
        return {}

    try:
        if fp.suffix.lower() == ".json":
            with fp.open("r", encoding="utf-8") as f:
                logger.debug("Reading cookies from JSON file: %s", file_path)
                cookies = json.load(f)
                if not isinstance(cookies, dict):
                    logger.error(
                        "Invalid JSON format. Expected a dictionary in file: %s",
                        file_path,
                    )
                    return {}
                return {str(k): str(v) for k, v in cookies.items()}
        elif fp.suffix.lower() == ".txt":
            logger.debug("Reading cookies from MozillaCookieJar file: %s", file_path)
            cookie_jar = MozillaCookieJar(fp)
            cookie_jar.load(ignore_discard=True, ignore_expires=True)
            return {cookie.name: cookie.value for cookie in cookie_jar}
        else:
            logger.error("Unsupported file type: %s. Supported types: .json, .txt", fp.suffix)
            return {}
    except json.JSONDecodeError as e:
        logger.error("JSON decoding error in file %s: %s", file_path, e)
    except OSError as e:
        logger.error("OS error when accessing file %s: %s", file_path, e)
    except Exception as e:
        logger.error("Unexpected error: %s", e)

    return {}



import queue
import asyncio
import threading
from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from logging import Logger
from queue import Queue
from typing import Any


@dataclass
class Task:
    """Unified task container for both threading and async services."""

    task_id: str
    func: Callable[..., Any]
    args: tuple[Any, ...] = ()
    kwargs: dict[str, Any] | None = None

    def __post_init__(self) -> None:
        self.kwargs = self.kwargs or {}


class BaseTaskService(ABC):
    """Abstract base class for task processing services."""

    @abstractmethod
    def start(self) -> None:
        """Start the service."""
        pass

    @abstractmethod
    def add_task(self, task: Task) -> None:
        """Add single task to queue."""
        pass

    @abstractmethod
    def add_tasks(self, tasks: list[Task]) -> None:
        """Add multiple tasks to queue."""
        pass

    @abstractmethod
    def get_result(self, task_id: str) -> Any | None:
        """Get result for specific task ID."""
        pass

    @abstractmethod
    def get_results(self, max_results: int = 0) -> dict[str, Any]:
        """Get all available results."""
        pass

    @abstractmethod
    def stop(self, timeout: int | None = None) -> None:
        """Stop the service."""
        pass


class ThreadingService(BaseTaskService):
    """Service for processing tasks with multiple workers."""

    def __init__(self, logger: Logger, max_workers: int = 5):
        self.task_queue: Queue[Task | None] = Queue()
        self.logger = logger
        self.max_workers = max_workers
        self.workers: list[threading.Thread] = []
        self.results: dict[str, Any] = {}
        self._lock = threading.Lock()
        self.is_running = False

    def start(self) -> None:
        if not self.is_running:
            self.is_running = True
            for _ in range(self.max_workers):
                worker = threading.Thread(target=self._process_tasks, daemon=True)
                self.workers.append(worker)
                worker.start()

    def _process_tasks(self) -> None:
        while True:
            task = self.task_queue.get()
            if task is None:
                break

            try:
                result = task.func(*task.args, **task.kwargs)  # type: ignore
                with self._lock:
                    self.results[task.task_id] = result
            except Exception as e:
                self.logger.error("Error processing task %s: %s", task.task_id, e)
            finally:
                self.task_queue.task_done()

    def add_task(self, task: Task) -> None:
        self.task_queue.put(task)
        if not self.is_running:
            self.start()

    def add_tasks(self, tasks: list[Task]) -> None:
        for task in tasks:
            self.task_queue.put(task)
        if not self.is_running:
            self.start()

    def get_result(self, task_id: str) -> Any | None:
        with self._lock:
            return self.results.pop(task_id, None)

    def get_results(self, max_results: int = 0) -> dict[str, Any]:
        with self._lock:
            if max_results <= 0:
                results_to_return = self.results.copy()
                self.results.clear()
                return results_to_return

            keys = list(self.results.keys())[:max_results]
            return {key: self.results.pop(key) for key in keys}

    def stop(self, timeout: int | None = None) -> None:
        self.task_queue.join()
        for _ in range(self.max_workers):
            self.task_queue.put(None)
        for worker in self.workers:
            worker.join(timeout=timeout)
        self.workers.clear()
        self.is_running = False


class AsyncService(BaseTaskService):
    def __init__(self, logger: Logger, max_workers: int = 5) -> None:
        self.max_workers = max_workers
        self.logger = logger
        self.is_running = False
        self.loop: asyncio.AbstractEventLoop | None = None
        self.thread: threading.Thread | None = None
        self._lock = threading.Lock()

        self.task_queue: queue.Queue[Task] = queue.Queue()
        self.results: dict[str, Any] = {}
        self.current_tasks: list[asyncio.Task[Any]] = []
        self.sem = asyncio.Semaphore(self.max_workers)

    def start(self) -> None:
        self._check_thread()

    def add_task(self, task: Task) -> None:
        self.task_queue.put(task)
        self._check_thread()

    def add_tasks(self, tasks: list[Task]) -> None:
        for task in tasks:
            self.task_queue.put(task)
        self._check_thread()

    def get_result(self, task_id: str) -> Any | None:
        with self._lock:
            return self.results.pop(task_id, None)

    def get_results(self, max_results: int = 0) -> dict[str, Any]:
        with self._lock:
            if max_results <= 0:
                results_to_return = self.results.copy()
                self.results.clear()
                return results_to_return

            keys = list(self.results.keys())[:max_results]
            return {key: self.results.pop(key) for key in keys}

    async def _process_tasks(self) -> None:
        while True:
            self.current_tasks = [task for task in self.current_tasks if not task.done()]

            if self.task_queue.empty() and not self.current_tasks:
                break

            while not self.task_queue.empty() and len(self.current_tasks) < self.max_workers:
                try:
                    task = self.task_queue.get_nowait()
                    task_obj = asyncio.create_task(self._run_task(task))
                    self.current_tasks.append(task_obj)
                except queue.Empty:
                    break

            if self.current_tasks:
                await asyncio.wait(self.current_tasks, return_when=asyncio.FIRST_COMPLETED)

    async def _run_task(self, task: Task) -> Any:
        async with self.sem:
            try:
                result = await task.func(*task.args, **task.kwargs)  # type: ignore
                with self._lock:
                    self.results[task.task_id] = result
                return result
            except Exception as e:
                self.logger.error(f"Error processing task {task.task_id}: {e}")
                with self._lock:
                    self.results[task.task_id] = None

    def _check_thread(self) -> None:
        with self._lock:
            if not self.is_running or self.thread is None or not self.thread.is_alive():
                self.is_running = True
                self.thread = threading.Thread(target=self._start_event_loop)
                self.thread.start()

    def _start_event_loop(self) -> None:
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        try:
            self.loop.run_until_complete(self._process_tasks())
        finally:
            self.loop.close()
            self.loop = None
            self.is_running = False
            self.current_tasks.clear()

    def stop(self, timeout: int | None = None) -> None:
        if self.thread is not None:
            self.thread.join(timeout=timeout)
            self.thread = None
            self.is_running = False



import os
import time
import random
from abc import ABC, abstractmethod
from logging import Logger
from subprocess import run
from typing import Any

from ..common import BaseConfig, RuntimeConfig, const
from ..utils import AccountManager, KeyManager


class BaseBot(ABC):
    """Abstract base class for bots, defining shared behaviors."""

    def __init__(
        self,
        runtime_config: RuntimeConfig,
        base_config: BaseConfig,
        key_manager: KeyManager,
        account_manager: AccountManager,
    ):
        self.runtime_config = runtime_config
        self.base_config = base_config
        self.close_browser = runtime_config.terminate
        self.logger = runtime_config.logger

        self.key_manager = key_manager
        self.account_manager = account_manager
        key_pair = self.key_manager.load_keys()
        self.private_key, self.public_key = key_pair.private_key, key_pair.public_key

        self.new_profile = False

    @abstractmethod
    def init_driver(self) -> Any:
        """Initialize the browser driver."""

    @abstractmethod
    def close_driver(self) -> None:
        """Close the browser and handle cleanup."""

    def prepare_chrome_profile(self) -> str:
        user_data_dir = self.base_config.chrome.profile_path

        if not os.path.exists(user_data_dir):
            os.makedirs(user_data_dir)
            self.new_profile = True
        else:
            self.new_profile = False

        return user_data_dir

    def auto_page_scroll(
        self,
        url: str,
        max_retry: int = 3,
        page_sleep: int = 5,
        fast_scroll: bool = True,
    ) -> str:
        """Scroll page automatically with retries and Cloudflare challenge handle.

        The main function of this class.

        Args:
            url (str): Target URL
            max_retry (int): Maximum number of retry attempts. Defaults to 3
            page_sleep (int): The sleep time after reaching page bottom
            fast_scroll (bool): Whether to use fast scroll. Might be blocked by Cloudflare

        Returns:
            str: Page HTML content or error message
        """
        raise NotImplementedError("Subclasses must implement automated retry logic.")

    def handle_login(self) -> None:
        """Login logic, implemented by subclasses."""
        raise NotImplementedError("Subclasses must implement login logic.")

    def human_like_type(self, element: Any, text: str) -> None:
        """Simulate human-like typing into a field."""
        raise NotImplementedError("Subclasses must implement scroll behavior.")

    def scroll_page(self) -> None:
        """Simulate human-like scrolling behavior."""
        raise NotImplementedError("Subclasses must implement scroll behavior.")


class BaseBehavior:
    pause_time = (0.1, 0.3)

    @staticmethod
    def random_sleep(min_time: float = 1.0, max_time: float = 5.0) -> None:
        time.sleep(random.uniform(min_time, max_time))


class BaseScroll:
    def __init__(self, base_config: BaseConfig, logger: Logger) -> None:
        self.base_config = base_config
        self.logger = logger
        self.scroll_position = 0
        self.last_content_height = 0
        self.successive_scroll_count = 0
        self.max_successive_scrolls = random.randint(5, 10)


def get_chrome_version_unix(chrome_path: str) -> str:
    try:
        result = run([chrome_path, "--version"], capture_output=True, text=True)
        if result.returncode == 0:
            return result.stdout.strip().split()[-1]
    except Exception:
        pass
    return const.DEFAULT_VERSION


def get_chrome_version() -> str:
    system = const.USER_OS

    if system == "Windows":
        import winreg

        try:
            reg_path = r"SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\chrome.exe"
            with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, reg_path) as key:  # type: ignore
                chrome_path, _ = winreg.QueryValueEx(key, "")  # type: ignore

            if os.path.exists(chrome_path):
                chrome_path = chrome_path.replace("\\", "\\\\")
                file_info = os.popen(  # nosec
                    f'wmic datafile where name="{chrome_path}" get Version /value',
                ).read()
                version = [
                    line.split("=")[1] for line in file_info.splitlines() if "Version=" in line
                ]
                return version[0] if version else const.DEFAULT_VERSION
        except Exception:
            pass
        return const.DEFAULT_VERSION

    if system == "Darwin":  # macOS
        return get_chrome_version_unix(const.DEFAULT_CONFIG["chrome"]["exec_path"]["Darwin"])

    if system == "Linux":
        return get_chrome_version_unix(const.DEFAULT_CONFIG["chrome"]["exec_path"]["Linux"])

    return const.DEFAULT_VERSION



import time
from typing import Any

from .drission_bot import DrissionBot
from .selenium_bot import SeleniumBot
from ..common import BaseConfig, RuntimeConfig
from ..utils import AccountManager, KeyManager


def get_bot(runtime_config: RuntimeConfig, app_config: BaseConfig) -> Any:
    bot_classes = {"selenium": SeleniumBot, "drission": DrissionBot}

    bot_type = runtime_config.bot_type
    logger = runtime_config.logger
    key_manager = KeyManager(logger, app_config.encryption)
    account_manager = AccountManager(logger, key_manager)

    if bot_type not in bot_classes:
        raise ValueError(f"Unsupported automator type: {bot_type}")

    bot = bot_classes[bot_type](runtime_config, app_config, key_manager, account_manager)

    if bot.new_profile:
        init_new_profile(bot)
    return bot


def init_new_profile(bot: Any) -> None:
    # visit some websites for new chrome profile
    websites: list[str] = [
        # "https://www.google.com",
        # "https://www.youtube.com",
        # "https://www.wikipedia.org",
    ]

    for url in websites:
        if isinstance(bot, DrissionBot):
            bot.page.get(url)
        elif isinstance(bot, SeleniumBot):
            bot.driver.get(url)

        time.sleep(4)



import re
import time
from abc import ABC, abstractmethod
from typing import Any, ClassVar, Generic, Literal, TypeAlias, TypeVar

from lxml import html

from ..common import BaseConfig, RuntimeConfig, ScrapeError
from ..common.const import BASE_URL
from ..utils import AlbumTracker, LinkParser, Task

# Manage return types of each scraper here
AlbumLink: TypeAlias = str
ImageLinkAndALT: TypeAlias = tuple[str, str]
LinkType = TypeVar("LinkType", AlbumLink, ImageLinkAndALT)
ScrapeType = Literal["album_list", "album_image"]


class ScrapeManager:
    """Manage the starting and ending of the scraper."""

    def __init__(
        self,
        runtime_config: RuntimeConfig,
        base_config: BaseConfig,
        web_bot: Any,
    ) -> None:
        self.runtime_config = runtime_config
        self.base_config = base_config

        self.web_bot = web_bot
        self.dry_run = runtime_config.dry_run
        self.logger = runtime_config.logger

        self.download_service = runtime_config.download_service

    def start_scraping(self) -> None:
        """Start scraping based on URL type."""
        try:
            urls = self._load_urls()
            for url in urls:
                self.runtime_config.url = url
                link_scraper = ScrapeHandler(self.runtime_config, self.base_config, self.web_bot)
                link_scraper.scrape(url, self.dry_run)
        except ScrapeError as e:
            self.logger.exception("Scraping error: '%s'", e)
        finally:
            self.download_service.stop()  # DO NOT REMOVE
            self.web_bot.close_driver()

    def _load_urls(self) -> list[str]:
        """Load URLs from runtime_config (URL or txt file)."""
        if self.runtime_config.input_file:
            with open(self.runtime_config.input_file) as file:
                urls = [line.strip() for line in file if line.strip()]
        else:
            urls = [self.runtime_config.url]
        return urls


class ScrapeHandler:
    """Handles all scraper behaviors."""

    # Defines the mapping from url part to scrape method.
    URL_HANDLERS: ClassVar[dict[str, ScrapeType]] = {
        "album": "album_image",
        "actor": "album_list",
        "company": "album_list",
        "category": "album_list",
        "country": "album_list",
    }

    def __init__(
        self,
        runtime_config: RuntimeConfig,
        base_config: BaseConfig,
        web_bot: Any,
    ) -> None:
        self.web_bot = web_bot
        self.logger = runtime_config.logger
        self.runtime_config = runtime_config

        self.strategies: dict[ScrapeType, BaseScraper[Any]] = {
            "album_list": AlbumScraper(
                runtime_config,
                base_config,
                web_bot,
                runtime_config.download_function,
            ),
            "album_image": ImageScraper(
                runtime_config,
                base_config,
                web_bot,
                runtime_config.download_function,
            ),
        }

        self.album_tracker = AlbumTracker(base_config.paths.download_log)
        self.path_parts, self.start_page = LinkParser.parse_input_url(runtime_config.url)

    def scrape(self, url: str, dry_run: bool = False) -> None:
        """Main entry point for scraping operations."""
        scrape_type = self._get_scrape_type()
        if scrape_type == "album_list":
            self.scrape_album_list(url, self.start_page, dry_run)
        else:
            self.scrape_album(url, self.start_page, dry_run)

    def scrape_album_list(self, url: str, start_page: int, dry_run: bool) -> None:
        """Handle scraping of album lists."""
        album_links = self._real_scrape(url, start_page, "album_list")
        self.logger.info("Found %d albums", len(album_links))

        for album_url in album_links:
            if dry_run:
                self.logger.info("[DRY RUN] Album URL: %s", album_url)
            else:
                self.scrape_album(album_url, 1, dry_run)

    def scrape_album(self, album_url: str, start_page: int, dry_run: bool) -> None:
        """Handle scraping of a single album page."""
        if self.album_tracker.is_downloaded(album_url) and not self.runtime_config.no_skip:
            self.logger.info("Album %s already downloaded, skipping.", album_url)
            return

        image_links = self._real_scrape(album_url, start_page, "album_image")
        if not image_links:
            return

        album_name = re.sub(r"\s*\d+$", "", image_links[0][1])
        self.logger.info("Found %d images in album %s", len(image_links), album_name)

        if dry_run:
            for link, _ in image_links:
                self.logger.info("[DRY RUN] Image URL: %s", link)
        else:
            self.album_tracker.log_downloaded(album_url)

    def _real_scrape(
        self,
        url: str,
        start_page: int,
        scrape_type: ScrapeType,
        **kwargs: dict[Any, Any],
    ) -> list[Any]:
        """Scrapes pages for links using the specified scraping strategy.

        Args:
            url (str): The URL to scrape.
            start_page (int): The starting page number for the scraping process.
            scrape_type (ScrapeType): The type of content to scrape, either "album" or "album_list".
            **kwargs (dict[Any, Any]): Additional keyword arguments for custom behavior.

        Returns:
            list[Any]: A list of results extracted from all scraped pages.

        Raises:
            KeyError: If the provided scrape_type is not found in the strategies.
        """
        strategy = self.strategies[scrape_type]
        self.logger.info(
            "Starting to scrape %s links from %s",
            "album" if scrape_type else "image",
            url,
        )

        all_results: list[Any] = []
        page = start_page

        while True:
            page_results, should_continue = self._scrape_single_page(
                url,
                page,
                strategy,
                scrape_type,
            )
            all_results.extend(page_results)

            if not should_continue:
                break

            page = self._handle_pagination(page)

        return all_results

    def _scrape_single_page(
        self,
        url: str,
        page: int,
        strategy: "BaseScraper[Any]",
        scrape_type: ScrapeType,
    ) -> tuple[list[AlbumLink] | list[ImageLinkAndALT], bool]:
        """Scrapes a single page and retrieves results with a flag indicating whether to continue scraping.

        Args:
            url (str): The URL to scrape.
            page (int): The page number to scrape.
            strategy (BaseScraper[Any]): The scraping strategy that defines how to extract data from the page.
            scrape_type (ScrapeType): The type of content to scrape, either "album" or "album_list".

        Returns:
            tuple[list[AlbumLink] | list[ImageLinkAndALT], bool]: A tuple containing:
            - list[AlbumLink] | list[ImageLinkAndALT]: A list of links or image details extracted from the page.
            - bool: A flag indicating whether to continue to the next page.
        """
        full_url = LinkParser.add_page_num(url, page)
        html_content = self.web_bot.auto_page_scroll(full_url, page_sleep=0)
        tree = LinkParser.parse_html(html_content, self.logger)

        if tree is None:
            return [], False

        self.logger.info("Fetching content from %s", full_url)
        page_links = tree.xpath(strategy.get_xpath())

        if not page_links:
            self.logger.info(
                "No more %s found on page %d",
                "albums" if scrape_type == "album_list" else "images",
                page,
            )
            return [], False

        page_result: list[AlbumLink] | list[ImageLinkAndALT] = []
        strategy.process_page_links(page_links, page_result, tree, page)

        # Check if we've reached the last page
        should_continue = page < LinkParser.get_max_page(tree)
        if not should_continue:
            self.logger.info("Reach last page, stopping")

        return page_result, should_continue

    def _handle_pagination(
        self,
        current_page: int,
        max_consecutive_page: int = 3,
        consecutive_sleep: int = 15,
    ) -> int:
        """Handle pagination logic including sleep for consecutive pages."""
        next_page = current_page + 1
        if next_page % max_consecutive_page == 0:
            time.sleep(consecutive_sleep)
        return next_page

    def _get_scrape_type(self) -> ScrapeType:
        """Get the appropriate handler method based on URL path."""
        for part in self.path_parts:
            if part in self.URL_HANDLERS:
                return self.URL_HANDLERS[part]
        raise ValueError(f"Unsupported URL type: {self.runtime_config.url}")


class BaseScraper(Generic[LinkType], ABC):
    """Abstract base class for different scraping strategies."""

    def __init__(
        self,
        runtime_config: RuntimeConfig,
        base_config: BaseConfig,
        web_bot: Any,
        download_function: Any,
    ) -> None:
        self.runtime_config = runtime_config
        self.base_config = base_config
        self.web_bot = web_bot
        self.download_service = runtime_config.download_service
        self.logger = runtime_config.logger

    @abstractmethod
    def get_xpath(self) -> str:
        """Return xpath for the specific strategy."""

    @abstractmethod
    def process_page_links(
        self,
        page_links: list[str],
        page_result: list[LinkType],
        tree: html.HtmlElement,
        page: int,
        **kwargs: dict[Any, Any],
    ) -> None:
        """Process links found on the page."""


class AlbumScraper(BaseScraper[AlbumLink]):
    """Strategy for scraping album list pages."""

    XPATH_ALBUM_LIST = '//a[@class="media-cover"]/@href'

    def get_xpath(self) -> str:
        return self.XPATH_ALBUM_LIST

    def process_page_links(
        self,
        page_links: list[str],
        page_result: list[AlbumLink],
        tree: html.HtmlElement,
        page: int,
        **kwargs: dict[Any, Any],
    ) -> None:
        page_result.extend([BASE_URL + album_link for album_link in page_links])
        self.logger.info("Found %d albums on page %d", len(page_links), page)


class ImageScraper(BaseScraper[ImageLinkAndALT]):
    """Strategy for scraping album image pages."""

    XPATH_ALBUM = '//div[@class="album-photo my-2"]/img/@data-src'
    XPATH_ALTS = '//div[@class="album-photo my-2"]/img/@alt'

    def __init__(
        self,
        runtime_config: RuntimeConfig,
        base_config: BaseConfig,
        web_bot: Any,
        download_function: Any,
    ) -> None:
        super().__init__(runtime_config, base_config, web_bot, download_function)
        self.dry_run = runtime_config.dry_run
        self.download_function = download_function
        self.alt_counter = 0

    def get_xpath(self) -> str:
        return self.XPATH_ALBUM

    def process_page_links(
        self,
        page_links: list[str],
        page_result: list[ImageLinkAndALT],
        tree: html.HtmlElement,
        page: int,
        **kwargs: dict[Any, Any],
    ) -> None:
        alts: list[str] = tree.xpath(self.XPATH_ALTS)

        # Handle missing alt texts
        # if len(alts) < len(page_links):
        #     missing_alts = [str(i + self.alt_counter) for i in range(len(page_links) - len(alts))]
        #     alts.extend(missing_alts)
        #     self.alt_counter += len(missing_alts)

        page_result.extend(zip(page_links, alts, strict=False))
        _, page_num = LinkParser.parse_input_url(self.runtime_config.url)
        idx = (page_num - 1) * 10 + 1

        # Handle downloads if not in dry run mode
        if not self.dry_run:
            album_name = extract_album_name(alts)

            # assign download job for each image
            for i, (url, _) in enumerate(zip(page_links, alts, strict=False)):
                task_id = f"{album_name}_{i}"
                task = Task(
                    task_id=task_id,
                    func=self.download_function,
                    kwargs={
                        "album_name": task_id,
                        "url": url,
                        "filename": f"{(idx + i):03d}",
                        "base_folder": self.base_config.download.download_dir,
                    },
                )
                self.download_service.add_task(task)

        self.logger.info("Found %d images on page %d", len(page_links), page)


def extract_album_name(alts: list[str]) -> str:
    album_name = next((alt for alt in alts if not alt.isdigit()), None)
    if album_name:
        album_name = re.sub(r"\s*\d*$", "", album_name).strip()
    if not album_name:
        album_name = BASE_URL.rstrip("/").split("/")[-1]
    return album_name



import os
import platform
from pathlib import Path
from typing import Any

import yaml
from dotenv import load_dotenv

from ._types import BaseConfig, ChromeConfig, DownloadConfig, EncryptionConfig, PathConfig


class ConfigPathTool:
    @staticmethod
    def resolve_abs_path(path: str | Path, base_dir: str | Path) -> str | Path:
        """Resolve '~', add path with base_dir if input is not absolute path."""
        path = os.path.expanduser(path)
        return os.path.join(base_dir, path) if not os.path.isabs(path) else path

    @staticmethod
    def get_system_config_dir() -> Path:
        """Return the config directory."""
        if platform.system() == "Windows":
            base = os.getenv("APPDATA", "")
        else:
            base = os.path.expanduser("~/.config")
        return Path(base) / "v2dl"

    @staticmethod
    def get_default_download_dir() -> Path:
        return Path.home() / "Downloads"

    @staticmethod
    def get_download_dir(download_dir: str) -> str:
        sys_dl_dir = ConfigPathTool.get_default_download_dir()
        result_dir = (
            ConfigPathTool.resolve_abs_path(download_dir, sys_dl_dir)
            if download_dir
            else sys_dl_dir
        )
        result_dir = Path(result_dir)
        result_dir.mkdir(parents=True, exist_ok=True)
        return str(result_dir)

    @staticmethod
    def get_chrome_exec_path(config_data: dict[str, Any]) -> str:
        current_os = platform.system()
        exec_path = config_data["chrome"]["exec_path"].get(current_os)
        if not exec_path:
            raise ValueError(f"Unsupported OS: {current_os}")
        if not isinstance(exec_path, str):
            raise TypeError(f"Expected a string for exec_path, got {type(exec_path).__name__}")
        return exec_path


class BaseConfigManager(ConfigPathTool):
    """Load and process configs based on user platform.

    The DEFAULT_CONFIG is a nested dict, after processing, the ConfigManager.load() returns a
    Config dataclass consists of DownloadConfig, PathConfig, ChromeConfig dataclasses.
    """

    def __init__(self, base_config: dict[str, dict[str, Any]], config_dir: str | None = None):
        self.base_config = base_config
        self.config_dir = config_dir

    def load(self) -> BaseConfig:
        """Load configuration from files and environment."""
        system_config_dir = BaseConfigManager.get_system_config_dir()
        if self.config_dir is not None:  # overwrite the config_dir
            system_config_dir = Path(self.config_dir)
        system_config_dir.mkdir(parents=True, exist_ok=True)

        custom_config_path = system_config_dir / "config.yaml"
        custom_env_path = system_config_dir / ".env"

        # Load environment variables
        if custom_env_path.exists():
            load_dotenv(custom_env_path)

        # Load and merge configurations
        if custom_config_path.exists():
            with open(custom_config_path) as f:
                custom_config = yaml.safe_load(f)
                if custom_config:  # not empty
                    self.base_config = BaseConfigManager._merge_config(
                        self.base_config,
                        custom_config,
                    )

        # Check file paths
        for key, path in self.base_config["paths"].items():
            self.base_config["paths"][key] = BaseConfigManager.resolve_abs_path(
                path,
                system_config_dir,
            )

        self.base_config["chrome"]["profile_path"] = BaseConfigManager.resolve_abs_path(
            self.base_config["chrome"]["profile_path"],
            system_config_dir,
        )

        # Check download_dir path
        download_dir = self.base_config["download"].get("download_dir", "").strip()
        self.base_config["download"]["download_dir"] = BaseConfigManager.get_download_dir(
            download_dir,
        )

        return BaseConfig(
            download=DownloadConfig(**self.base_config["download"]),
            paths=PathConfig(**self.base_config["paths"]),
            chrome=ChromeConfig(
                exec_path=BaseConfigManager.get_chrome_exec_path(self.base_config),
                profile_path=self.base_config["chrome"]["profile_path"],
            ),
            encryption=EncryptionConfig(
                **self.base_config.get("encryption", self.base_config["encryption"]),
            ),
        )

    @staticmethod
    def _merge_config(base: dict[str, Any], custom: dict[str, Any]) -> dict[str, Any]:
        """Recursively merge custom config into base config."""
        for key, value in custom.items():
            if isinstance(value, dict) and key in base:
                BaseConfigManager._merge_config(base[key], value)
            else:
                base[key] = value
        return base



class ScrapeError(Exception):
    """All scraping related errors."""


class FileProcessingError(ScrapeError):
    """File processing fail."""


class DownloadError(ScrapeError):
    """Downloading fail."""


class SecurityError(Exception):
    """Password encryption/decryption fail."""


class BotError(Exception):
    """Web bot operation error."""



import sys
import time
import random
from logging import Logger
from typing import TYPE_CHECKING, Any

from DrissionPage import ChromiumOptions, ChromiumPage
from DrissionPage.common import wait_until
from DrissionPage.errors import ElementNotFoundError, WaitTimeoutError

from .base import BaseBehavior, BaseBot, BaseScroll
from .cookies import load_cookies
from ..common.error import BotError

if TYPE_CHECKING:
    from ..common import BaseConfig, RuntimeConfig
    from ..utils import AccountManager, KeyManager


class DrissionBot(BaseBot):
    def __init__(
        self,
        runtime_config: "RuntimeConfig",
        base_config: "BaseConfig",
        key_manager: "KeyManager",
        account_manager: "AccountManager",
    ) -> None:
        super().__init__(runtime_config, base_config, key_manager, account_manager)
        self.base_config = base_config
        self.init_driver()
        self.cloudflare = DriCloudflareHandler(self.page, self.logger)

    def init_driver(self) -> None:
        co = ChromiumOptions()
        if self.runtime_config.chrome_args is not None:
            for conf in self.runtime_config.chrome_args:
                co.set_argument(conf)

        if self.runtime_config.user_agent is not None:
            co.set_user_agent(user_agent=self.runtime_config.user_agent)

        if not self.runtime_config.use_chrome_default_profile:
            user_data_dir = self.prepare_chrome_profile()
            co.set_user_data_path(user_data_dir)
        else:
            co.use_system_user_path()

        self.page = ChromiumPage(addr_or_opts=co, timeout=0.8)  # type: ignore
        self.page.set.scroll.smooth(on_off=True)
        self.page.set.scroll.wait_complete(on_off=True)

        self.scroller = DriScroll(self.page, self.base_config, self.logger)

    def close_driver(self) -> None:
        if self.close_browser:
            self.page.quit()

    def auto_page_scroll(
        self,
        url: str,
        max_retry: int = 3,
        page_sleep: int = 5,
        fast_scroll: bool = False,
    ) -> str:
        scroll_down = self.page.scroll.to_bottom if fast_scroll else self.scroller.scroll_to_bottom

        for attempt in range(max_retry):
            try:
                self.page.get(url)

                # handle page redirection fail
                if not self.handle_redirection_fail(url, max_retry, page_sleep):
                    self.logger.error(
                        "Reconnection fail for URL %s. Please check your network status.",
                        url,
                    )
                    break

                if self.cloudflare.handle_simple_block(attempt, max_retry):
                    continue

                # main business
                self.handle_login()
                self.handle_read_limit()
                self.page.run_js("document.body.style.zoom='75%'")
                scroll_down()

                # Sleep to avoid Cloudflare blocking
                self.logger.debug("Scrolling finished, pausing to avoid blocking")
                DriBehavior.random_sleep(page_sleep, page_sleep + 5)
                break

            except Exception as e:
                self.logger.exception(
                    "Request failed for URL %s - Attempt %d/%d. Error: %s",
                    url,
                    attempt + 1,
                    max_retry,
                    e,
                )
                DriBehavior.random_sleep(page_sleep, page_sleep + 5)

        if not self.page.html:
            error_template = "Failed to retrieve URL after {} attempts: '{}'"
            error_msg = error_template.format(max_retry, url)
            self.logger.error(error_msg)
            return error_msg
        return self.page.html

    def handle_redirection_fail(self, url: str, max_retry: int, sleep_time: int) -> bool:
        # If read limit exceed, not a redirection fail.
        # If not exceed read limit, check url.
        if self.check_read_limit() or (self.page.url == url and self.page.states.is_alive):
            return True
        retry = 1
        while retry <= max_retry:
            self.logger.error(
                "Redirection handle failed for URL %s - Attempt %d/%d.",
                url,
                retry,
                max_retry,
            )
            DriBehavior.random_sleep(sleep_time, sleep_time + 5 * random.uniform(1, retry * 5))

            if self.cloudflare.handle_simple_block(retry, max_retry):
                self.logger.warning("Failed to solve Cloudflare turnstile challenge")
                continue

            self.page.get(url)
            retry += 1
            if self.page.url == url and self.page.states.is_alive:
                return True

        return self.page.url == url and self.page.states.is_alive

    def handle_login(self) -> None:
        success = False
        if self.page("x://h1[@class='h4 text-secondary mb-4 login-box-msg']"):
            self.logger.info("Login page detected - Starting login process")
            try:
                for _ in self.account_manager.accounts:
                    # this will update self.email and cookies_valid
                    # if no accounts available, `AccountManager.random_pick` will execute sys.exit
                    if self.cookies_login():
                        return
                    email_field = self.page("#email")
                    password_field = self.page("#password")

                    # self.handle_cloudflare_recaptcha()
                    email_field.clear(True)
                    password_field.clear(True)

                    DriBehavior.human_like_type(email_field, self.email)
                    DriBehavior.random_sleep(0.01, 0.3)
                    DriBehavior.human_like_type(password_field, self.password)
                    DriBehavior.random_sleep(0.01, 0.5)

                    login_button = self.page(
                        'x://button[@type="submit" and @class="btn btn-primary"]',
                    )
                    login_button.click()

                    if not self.page('x://a[@href="/site/recovery-password"]'):
                        self.logger.info("Account %s login successful with password", self.email)
                        success = True
                    else:
                        self.logger.info(
                            "Account %s login failed. Checking error messages",
                            self.email,
                        )
                        self.account_manager.update_runtime_state(
                            self.email,
                            "password_valid",
                            False,
                        )
                        self.check_login_errors()
                        return

            except ElementNotFoundError as e:
                self.logger.error("Login form element not found: %s", e)
                raise
            except WaitTimeoutError as e:
                self.logger.error("Timeout waiting for element: %s", e)
                raise
            except Exception as e:
                self.logger.error("Unexpected error during login: %s", e)
                raise

        else:
            success = True

        if not success:
            self.logger.info("Automated login failed. Please login yourself.")
            sys.exit("Automated login failed.")

    def cookies_login(self) -> bool:
        self.email, self.password = self.account_manager.random_pick(self.private_key)
        account = self.account_manager.read(self.email)
        if account is None:
            raise BotError("Unexpected error while reading account '%s'", self.email)

        cookies_path = account.get("cookies")
        if cookies_path:
            cookies = load_cookies(cookies_path)
            self.page.set.cookies.clear()
            self.page.set.cookies(cookies)
            self.page.refresh()

        if not self.page('x://a[@href="/site/recovery-password"]'):
            self.logger.info("Account %s login successful with cookies", self.email)
            return True

        self.account_manager.update_runtime_state(self.email, "cookies_valid", False)
        return False

    def check_login_errors(self) -> None:
        error_message = self.page('x://div[@class="errorMessage"]')
        if error_message:
            self.logger.error("Login error: %s", error_message.text)
        else:
            self.logger.error("Login failed for unknown reasons")
            sys.exit(1)

    def handle_read_limit(self) -> None:
        if self.check_read_limit():
            # click logout
            self.page('x://ul[@class="nav justify-content-end"]//a[@href="/user/logout"]').click()
            self.account_manager.update_account(self.email, "exceed_quota", True)
            self.email, self.password = self.account_manager.random_pick(self.private_key)

    def check_read_limit(self) -> bool:
        return self.page.url == "https://www.v2ph.com/user/upgrade"

    def click_logout(self) -> None:
        self.page.ele("@href=/user/logout").click()


class DriCloudflareHandler:
    """Handles Cloudflare protection detection and bypass attempts.

    Includes methods for dealing with various Cloudflare challenges.
    """

    def __init__(self, page: ChromiumPage, logger: Logger) -> None:
        self.page = page
        self.logger = logger

    def handle_simple_block(self, attempt: int, retries: int) -> bool:
        """Check, handle, and return whether blocked or not."""
        blocked = False
        if self.is_simple_blocked():
            self.logger.info(
                "Cloudflare challenge detected - Solve attempt %d/%d",
                attempt + 1,
                retries,
            )
            blocked = self.handle_cloudflare_turnstile()
        return blocked

    def handle_hard_block(self) -> bool:
        """Check whether blocked or not.

        This is a cloudflare WAF full page block.
        """
        blocked = False
        if self.is_hard_block():
            self.logger.warning("Hard block detected by Cloudflare - Unable to proceed")
            blocked = True
        return blocked

    def is_simple_blocked(self) -> bool:
        title_check = any(text in self.page.title for text in ["請稍候...", "Just a moment..."])
        page_source_check = "Checking your" in self.page.html
        return title_check or page_source_check

    def is_hard_block(self) -> bool:
        is_blocked = "Attention Required! | Cloudflare" in self.page.title
        if is_blocked:
            self.logger.warning("Cloudflare hard block detected")
        return is_blocked

    def handle_cloudflare_turnstile(self) -> bool:
        """鬥志鬥勇失敗."""
        blocked = False
        try:
            container = self.page.s_ele(".cloudflare-container")
            turnstile_box = container.s_ele(".turnstile-box")
            turnstile_div = turnstile_box.s_ele("#cf-turnstile")
            turnstile_div.rect.click_point()  # type: ignore
            self.page.wait(2)
            # pyautogui.moveTo(pos[0], pos[1] + 61, duration=0.5)
            # pyautogui.click()
            self.page.wait(3)
            blocked = True
        except Exception as e:
            self.logger.exception("Failed to solve new Cloudflare turnstile: %s", e)
        return blocked

    def random_sleep(self, min_time: float, max_time: float) -> None:
        time.sleep(random.uniform(min_time, max_time))


class DriBehavior(BaseBehavior):
    @staticmethod
    def human_like_mouse_movement(page: Any, element: Any) -> None:
        # Get the element's position
        rect = element.rect
        action_x = random.randint(-100, 100)
        action_y = random.randint(-100, 100)

        # Move by offset and then to element
        page.mouse.move_to(x=rect["x"] + action_x, y=rect["y"] + action_y)
        page.mouse.move_to(rect["x"], rect["y"])
        DriBehavior.random_sleep(*BaseBehavior.pause_time)

    @staticmethod
    def human_like_click(page: Any, element: Any) -> None:
        DriBehavior.human_like_mouse_movement(page, element)
        page.mouse.click()
        DriBehavior.random_sleep(*BaseBehavior.pause_time)

    @staticmethod
    def human_like_type(element: Any, text: str) -> None:
        for char in text:
            element.input(char)
            time.sleep(random.uniform(0.001, 0.2))
        DriBehavior.random_sleep(*BaseBehavior.pause_time)


class DriScroll(BaseScroll):
    def __init__(self, page: ChromiumPage, base_config: "BaseConfig", logger: Logger) -> None:
        super().__init__(base_config, logger)
        self.page = page
        self.page.set.scroll.smooth(on_off=True)

    def scroll_to_bottom(self) -> None:
        attempts = 0
        max_attempts = 10
        wait_time = (1, 2)
        last_position = -123459
        scroll_length = lambda: random.randint(
            self.base_config.download.min_scroll_length,
            self.base_config.download.max_scroll_length,
        )

        while attempts < max_attempts:
            scroll = scroll_length()
            self.logger.debug(
                "Current position: %d, scrolling down by %d pixels",
                last_position,
                scroll,
            )
            self.page.run_js(f"window.scrollBy({{top: {scroll}, behavior: 'smooth'}});")
            self.page.wait(*wait_time)

            new_position = self.page.run_js("return window.pageYOffset;")
            if new_position == last_position:
                break
            last_position = new_position
            attempts += 1

    def old_scroll_to_bottom(self) -> None:
        self.logger.info("Start scrolling the page")
        scroll_attempts = 0
        max_attempts = 45
        same_position_count = 0
        max_same_position_count = 3
        last_position = 0.0
        scrolled_up = False

        while scroll_attempts < max_attempts:
            scroll_attempts += 1

            current_position = self.get_current_position()

            if current_position == last_position:
                same_position_count += 1
                if same_position_count >= max_same_position_count:
                    self.logger.debug(
                        "Detected the same position three times in a row, stopping scrolling. Scrolled a total of %d times",
                        scroll_attempts,
                    )
                    break
            else:
                same_position_count = 0

            last_position = current_position

            scrolled_up = self.perform_scroll_action(scrolled_up)

            self.wait_for_content_load()

            self.successive_scroll_count += 1
            if self.successive_scroll_count >= self.max_successive_scrolls:
                pause_time = random.uniform(3, 7)
                self.logger.debug(
                    "Scrolled %d times, pausing for %.2f seconds",
                    self.successive_scroll_count,
                    pause_time,
                )
                time.sleep(pause_time)
                scrolled_up = False
                self.successive_scroll_count = 0
                self.max_successive_scrolls = random.randint(3, 7)

        if scroll_attempts == max_attempts:
            self.logger.info(
                "Reached maximum attempts (%d), scrolling finished, may not have fully reached the bottom",
                max_attempts,
            )
        else:
            self.logger.info("Page scroll completed")

    def perform_scroll_action(self, scrolled_up: bool) -> bool:
        while True:
            action = random.choices(
                ["scroll_down", "scroll_up", "pause", "jump"],
                weights=[0.9, 0.1, 0.1, 0.01],
            )[0]

            if (
                action != "scroll_up" or not scrolled_up
            ):  # 連續捲動時，只要往上捲動過一次就不要再選擇往上
                break

        if action == "scroll_down":
            scroll_length = random.randint(
                self.base_config.download.min_scroll_step,
                self.base_config.download.max_scroll_step,
            )
            self.logger.debug("Trying to scroll down %d", scroll_length)
            self.page.scroll.down(pixel=scroll_length)
            time.sleep(random.uniform(*BaseBehavior.pause_time))
        elif action == "scroll_up":
            scroll_length = random.randint(
                self.base_config.download.min_scroll_step,
                self.base_config.download.max_scroll_step,
            )
            self.logger.debug("Trying to scroll up %d", scroll_length)
            self.page.scroll.up(pixel=scroll_length)
        elif action == "pause":
            pause_time = random.uniform(1, 3)
            self.logger.debug("Pausing for %.2f seconds", pause_time)
            time.sleep(pause_time)
        elif action == "jump":
            self.logger.debug("Jumping to the bottom of the page")
            self.page.scroll.to_bottom()
            # self.page.scroll.to_see("@class=album-photo my-2")

        return action == "scrolled_up"

    # def safe_scroll(self, target_position):
    #     """DrissionPage does not need safe_scroll."""
    #     current_position = self.get_current_position()
    #     step = random.uniform(
    #         self.config.download.min_scroll_step,
    #         self.config.download.max_scroll_step,
    #     )

    #     while abs(current_position - target_position) > step:
    #         self.page.run_js(f"window.scrollTo(0, {current_position + step});")
    #         time.sleep(random.uniform(0.005, 0.1))
    #         new_position = self.get_current_position()
    #         if new_position == current_position:
    #             self.logger.debug(
    #                 "Cannot continue scrolling, target: %d, current: %d", target_position, current_position
    #             )
    #             break
    #         current_position = new_position
    #     self.page.run_js(f"window.scrollTo(0, {target_position});")
    #     return self.get_current_position()

    def get_current_position(self) -> float:
        page_location: float = self.page.run_js("return window.pageYOffset;")
        self.logger.debug("Current vertical position %d", page_location)
        return page_location

    def get_page_height(self) -> float:
        page_height: float = self.page.run_js("return document.body.scrollHeight;")
        self.logger.debug("Total page height %d", page_height)
        return page_height

    def wait_for_content_load(self) -> None:
        try:
            # wait until the callable return true. default time_out=10
            wait_until(lambda: self.page.states.ready_state == "complete", timeout=5)
        except TimeoutError:
            self.logger.warning("Timeout waiting for new content to load")



# ============== Default User Preference ==============
import platform
from typing import Any

DEFAULT_CONFIG: dict[str, dict[str, Any]] = {
    "download": {
        "min_scroll_length": 1000,
        "max_scroll_length": 2000,
        "min_scroll_step": 300,
        "max_scroll_step": 500,
        "rate_limit": 400,
        "download_dir": "v2dl",
    },
    "paths": {
        "download_log": "downloaded_albums.txt",
        "system_log": "v2dl.log",
    },
    "chrome": {
        "profile_path": "v2dl_chrome_profile",
        "exec_path": {
            "Linux": "/usr/bin/google-chrome",
            "Darwin": "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
            "Windows": r"C:\Program Files\Google\Chrome\Application\chrome.exe",
        },
    },
    "encryption": {
        "key_bytes": 32,
        "salt_bytes": 16,
        "nonce_bytes": 24,
        "kdf_ops_limit": 2**10,
        "kdf_mem_limit": 2**13,
    },
}


# ============== System ==============
BASE_URL = "https://www.v2ph.com"
DEMO_URL_ALBUM = "https://www.v2ph.com/album/Weekly-Young-Jump-2015-No15"
DEMO_URL_ALBUM_LIST = "https://www.v2ph.com/category/nogizaka46"
WORKFLOW_URL_ACTOR = (
    "https://www.v2ph.com/album/Weekly-Big-Comic-Spirits-2016-No22-23"  # only 1 page
)


# For selenium webdriver
USER_OS = platform.system()
DEFAULT_VERSION = "130.0.6723.59"
SELENIUM_AGENT = "user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{} Safari/537.36"


# For requests to download from the v2ph cdn, somehow the fake_useragent is not working.
HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.6723.59 Safari/537.36",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
    "Accept-Language": "en-US,en;q=0.9",
    "Accept-Encoding": "gzip, deflate, br",
    "Referer": "https://www.google.com/",
    "DNT": "1",
    "Connection": "keep-alive",
    "Upgrade-Insecure-Requests": "1",
    # "Sec-Fetch-Dest": "document",
    # "Sec-Fetch-Mode": "navigate",
    # "Sec-Fetch-Site": "none",
    # "Sec-Fetch-User": "?1",
    # "Cache-Control": "max-age=0",
}



import sys
import time
import random
from logging import Logger
from subprocess import Popen
from typing import TYPE_CHECKING, Any

from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from .base import BaseBehavior, BaseBot, BaseScroll, get_chrome_version
from .cookies import load_cookies
from ..common import SELENIUM_AGENT

if TYPE_CHECKING:
    from ..common import BaseConfig, RuntimeConfig
    from ..utils import AccountManager, KeyManager

DEFAULT_BOT_OPT = [
    "--remote-debugging-port=9222",
    "--disable-gpu",
    "--disable-infobars",
    "--disable-extensions",
    "--disable-dev-shm-usage",
    "--start-maximized",
]


class SeleniumBot(BaseBot):
    def __init__(
        self,
        runtime_config: "RuntimeConfig",
        base_config: "BaseConfig",
        key_manager: "KeyManager",
        account_manager: "AccountManager",
    ) -> None:
        super().__init__(runtime_config, base_config, key_manager, account_manager)
        self.init_driver()
        self.scroller = SelScroll(self.driver, self.base_config, self.logger)
        self.cloudflare = SelCloudflareHandler(self.driver, self.logger)

    def init_driver(self) -> None:
        self.driver: WebDriver
        options = Options()

        chrome_path = [self.base_config.chrome.exec_path]
        # commands for running subprocess
        subprocess_cmd = chrome_path + (self.runtime_config.chrome_args or DEFAULT_BOT_OPT)
        subprocess_cmd = [
            *subprocess_cmd,
            self.runtime_config.user_agent or SELENIUM_AGENT.format(get_chrome_version()),
        ]

        if not self.runtime_config.use_chrome_default_profile:
            user_data_dir = self.prepare_chrome_profile()
            subprocess_cmd.append(f"--user-data-dir={user_data_dir}")

        for arg in subprocess_cmd:
            options.add_argument(arg)

        # additional args for webdriver.Chrome to takeover the control of created browser
        options.add_experimental_option("debuggerAddress", "127.0.0.1:9222")
        try:
            self.chrome_process = Popen(subprocess_cmd)  # subprocess.run fails
            self.driver = webdriver.Chrome(service=Service(), options=options)
        except Exception as e:
            self.logger.error("Unable to start Selenium WebDriver: %s", e)
            sys.exit("Unable to start Selenium WebDriver")

    def close_driver(self) -> None:
        if self.close_browser:
            self.driver.quit()
            self.chrome_process.terminate()

    def auto_page_scroll(
        self,
        url: str,
        max_retry: int = 3,
        page_sleep: int = 5,
        fast_scroll: bool = False,
    ) -> str:
        response: str = ""
        for attempt in range(max_retry):
            try:
                self.driver.get(url)
                SelBehavior.random_sleep(0.1, 0.5)

                if not self.handle_redirection_fail(url, max_retry, page_sleep):
                    self.logger.error(
                        "Reconnection fail for URL %s. Please check your network status.",
                        url,
                    )
                    break

                if self.cloudflare.handle_simple_block(attempt, max_retry):
                    continue

                # main business
                self.handle_login()
                self.driver.execute_script("document.body.style.zoom='75%'")
                self.scroller.scroll_to_bottom()
                SelBehavior.random_sleep(5, 15)

                response = self.driver.page_source
                break

            except Exception as e:
                self.logger.exception(
                    "Request failed for URL %s - Attempt %d/%d. Error: %s",
                    url,
                    attempt + 1,
                    max_retry,
                    e,
                )

            self.logger.debug("Scrolling finished, pausing to avoid blocking")
            SelBehavior.random_sleep(page_sleep, page_sleep + 5)

        if not response:
            error_template = "Failed to retrieve URL after {} attempts: '{}'"
            error_msg = error_template.format(max_retry, url)
            self.logger.error(error_msg)
            return error_msg
        return response

    def handle_redirection_fail(self, url: str, max_retry: int, sleep_time: int) -> bool:
        if self.driver.current_url == url:
            return True
        retry = 1
        while retry <= max_retry:
            self.logger.error("Connection failed - Attempt %d/%d", retry, max_retry)
            SelBehavior.random_sleep(sleep_time, sleep_time + 5 * random.uniform(1, retry * 5))

            if self.cloudflare.handle_simple_block(retry, max_retry):
                self.logger.warning("Failed to solve Cloudflare turnstile challenge")
                continue

            self.driver.get(url)
            retry += 1
            if self.driver.current_url == url:
                return True

        return self.driver.current_url == url

    def handle_login(self) -> None:
        success = False
        if self.driver.find_elements(
            By.XPATH,
            "//h1[@class='h4 text-secondary mb-4 login-box-msg']",
        ):
            self.logger.info("Login page detected - Starting login process")
            try:
                for _ in self.account_manager.accounts:
                    # this will update self.email and cookies_valid
                    # if no accounts available, `AccountManager.random_pick` will execute sys.exit
                    if self.cookies_login():
                        return

                    self.email, self.password = self.account_manager.random_pick(self.private_key)
                    email_field = self.driver.find_element(By.ID, "email")
                    password_field = self.driver.find_element(By.ID, "password")
                    BaseBehavior.random_sleep(0.5, 1)

                    password_field.send_keys(Keys.SHIFT, Keys.TAB)
                    BaseBehavior.random_sleep(0.5, 1)
                    SelBehavior.human_like_type(email_field, self.email)
                    SelBehavior.random_sleep(0.01, 0.3)

                    email_field.send_keys(Keys.TAB)
                    SelBehavior.human_like_type(password_field, self.password)
                    SelBehavior.random_sleep(0.01, 0.5)

                    try:
                        self.cloudflare.handle_cloudflare_recaptcha()
                    except Exception as e:
                        self.logger.exception("Error handling Cloudflare reCAPTCHA: %s", e)

                    self.driver.find_element(
                        By.XPATH,
                        '//button[@type="submit" and @class="btn btn-primary"]',
                    ).click()

                    SelBehavior.random_sleep(3, 5)

                    if not self.driver.find_elements(
                        By.XPATH,
                        "//h1[@class='h4 text-secondary mb-4 login-box-msg']",
                    ):
                        self.logger.info("Login successful")
                        success = True
                    else:
                        self.logger.error("Login failed - Checking for error messages")
                        self.account_manager.update_runtime_state(
                            self.email,
                            "password_valid",
                            False,
                        )
                        self.check_login_errors()

            except NoSuchElementException as e:
                self.logger.error("Login form element not found: %s", e)
                raise
            except TimeoutException as e:
                self.logger.error("Timeout waiting for element: %s", e)
                raise
            except Exception as e:
                self.logger.error("Unexpected error during login: %s", e)
                raise
        else:
            success = True
        if not success:
            self.logger.info("Automated login failed. Please login yourself.")
            sys.exit("Automated login failed.")

    def cookies_login(self) -> bool:
        self.email, self.password = self.account_manager.random_pick(self.private_key)
        account = self.account_manager.read(self.email)
        if account is None:
            return False

        cookies_path = account.get("cookies")
        if cookies_path:
            cookies = load_cookies(cookies_path)
            self.driver.delete_all_cookies()
            for k, v in cookies.items():
                self.driver.add_cookie({"name": k, "value": v})
            self.driver.refresh()

        if not self.driver.find_element('//a[@href="/site/recovery-password"]'):
            self.logger.info("Account %s login successful with cookies", self.email)
            return True

        self.account_manager.update_runtime_state(self.email, "cookies_valid", False)
        return False

    def check_login_errors(self) -> None:
        error_messages = self.driver.find_elements(By.CLASS_NAME, "errorMessage")
        if error_messages:
            for message in error_messages:
                self.logger.error("Login error: %s", message.text)
        else:
            self.logger.warning(
                "No specific error message found. Login might have failed for unknown reasons.",
            )


class SelCloudflareHandler:
    """Handles Cloudflare protection detection and bypass attempts.

    Includes methods for dealing with various Cloudflare challenges.
    """

    def __init__(self, driver: WebDriver, logger: Logger):
        self.driver = driver
        self.logger = logger

    def handle_simple_block(self, attempt: int, retries: int) -> bool:
        """check and handle Cloudflare challenge."""
        blocked = False
        if self.is_simple_blocked():
            self.logger.info(
                "Detected Cloudflare challenge, attempting to solve... Attempt %d/%d",
                attempt + 1,
                retries,
            )
            self.handle_cloudflare_turnstile()
            blocked = True
        return blocked

    def handle_hard_block(self) -> bool:
        """Check whether blocked or not.

        This is a cloudflare WAF full page block.
        """
        blocked = False
        if self.is_hard_block():
            self.logger.warning("Hard block detected by Cloudflare - Unable to proceed")
            blocked = True
        return blocked

    def is_simple_blocked(self) -> bool:
        title_check = any(text in self.driver.title for text in ["請稍候...", "Just a moment..."])
        page_source_check = "Checking your" in self.driver.page_source
        return title_check or page_source_check

    def is_hard_block(self) -> bool:
        is_blocked = "Attention Required! | Cloudflare" in self.driver.title
        if is_blocked:
            self.logger.warning("Cloudflare hard block detected")
        return is_blocked

    def handle_cloudflare_turnstile(self) -> None:
        try:
            iframe = WebDriverWait(self.driver, 10).until(
                EC.presence_of_element_located((
                    By.XPATH,
                    "//iframe[contains(@src, 'challenges.cloudflare.com')]",
                )),
            )
            self.driver.switch_to.frame(iframe)

            checkbox = WebDriverWait(self.driver, 10).until(
                EC.element_to_be_clickable((By.ID, "cf-turnstile-response")),
            )
            SelBehavior.human_like_click(self.driver, checkbox)

            if "Select all squares with" in self.driver.page_source:
                self.solve_image_captcha()

            self.driver.switch_to.default_content()
            SelBehavior.random_sleep(10, 20)
        except (TimeoutException, NoSuchElementException):
            self.logger.error("Unable to solve Cloudflare challenge.")

    def handle_cloudflare_recaptcha(self) -> None:
        try:
            recaptcha_checkbox = WebDriverWait(self.driver, 10).until(
                EC.element_to_be_clickable((By.XPATH, "//input[@type='checkbox']")),
            )
            SelBehavior.human_like_click(self.driver, recaptcha_checkbox)

            verify_button = WebDriverWait(self.driver, 10).until(
                EC.element_to_be_clickable((By.XPATH, "//span[contains(text(), '驗證您是人類')]")),
            )
            SelBehavior.human_like_click(self.driver, verify_button)

            SelBehavior.random_sleep(3, 5)
        except (TimeoutException, NoSuchElementException) as e:
            self.logger.warning(
                "reCAPTCHA checkbox or verify button not found or unable to interact: %s",
                e,
            )

    def solve_image_captcha(self) -> None:
        raise NotImplementedError


class SelBehavior(BaseBehavior):
    @staticmethod
    def human_like_mouse_movement(driver: Any, element: Any) -> None:
        action = ActionChains(driver)
        action.move_by_offset(random.randint(-100, 100), random.randint(-100, 100))
        action.move_to_element_with_offset(
            element,
            random.randint(-10, 10),
            random.randint(-10, 10),
        )
        action.pause(random.uniform(0.1, 0.3))
        action.move_to_element(element)
        action.perform()
        SelBehavior.random_sleep(*BaseBehavior.pause_time)

    @staticmethod
    def human_like_click(driver: Any, element: Any) -> None:
        SelBehavior.human_like_mouse_movement(driver, element)
        action = ActionChains(driver)
        action.click()
        action.perform()
        SelBehavior.random_sleep(*BaseBehavior.pause_time)

    @staticmethod
    def human_like_type(element: Any, text: str) -> None:
        for char in text:
            element.send_keys(char)
            time.sleep(random.uniform(0.001, 0.2))
        SelBehavior.random_sleep(*BaseBehavior.pause_time)


class SelScroll(BaseScroll):
    def __init__(self, driver: WebDriver, base_config: "BaseConfig", logger: Logger):
        super().__init__(base_config, logger)
        self.driver = driver

    def scroll_to_bottom(self) -> None:
        max_attempts = 10
        attempts = 0
        last_position = -123459
        scroll_length = lambda: random.randint(
            self.base_config.download.min_scroll_length,
            self.base_config.download.max_scroll_length,
        )

        while attempts < max_attempts:
            SelBehavior.random_sleep(1, 2)
            scroll = scroll_length()
            self.logger.debug(
                "Current position: %d, scrolling down by %d pixels",
                last_position,
                scroll,
            )
            self.driver.execute_script(f"window.scrollBy({{top: {scroll}, behavior: 'smooth'}});")
            new_position = self.driver.execute_script("return window.pageYOffset;")
            if new_position == last_position:
                break
            last_position = new_position
            attempts += 1

    def old_scroll_to_bottom(self) -> None:
        self.logger.info("Start scrolling the page")
        scroll_attempts = 0
        max_attempts = 45

        scroll_pos_init = self.driver.execute_script("return window.pageYOffset;")
        step_scroll = random.randint(
            self.base_config.download.min_scroll_length,
            self.base_config.download.max_scroll_length,
        )

        while scroll_attempts < max_attempts:
            scroll_attempts += 1

            self.driver.execute_script(f"window.scrollBy(0, {step_scroll});")
            scroll_pos_end = self.driver.execute_script("return window.pageYOffset;")
            time.sleep(0.75)

            if scroll_pos_init >= scroll_pos_end:
                self.logger.debug("Reached the bottom of the page")
                break

            scroll_pos_init = scroll_pos_end

            step_scroll = random.randint(
                self.base_config.download.min_scroll_length,
                self.base_config.download.max_scroll_length,
            )

            self.wait_for_content_load()

            self.successive_scroll_count += 1
            if self.successive_scroll_count >= self.max_successive_scrolls:
                pause_time = random.uniform(3, 7)
                self.logger.debug(
                    "Scrolled %d times, pausing for %.2f seconds",
                    self.successive_scroll_count,
                    pause_time,
                )
                time.sleep(pause_time)
                self.successive_scroll_count = 0
                self.max_successive_scrolls = random.randint(3, 7)

        if scroll_attempts == max_attempts:
            self.logger.info(
                "Reached maximum attempts (%d), scrolling finished, may not have fully reached the bottom",
                max_attempts,
            )
        else:
            self.logger.info("Page scroll completed")

    def perform_scroll_action(self) -> None:
        action = random.choices(
            ["scroll_down", "scroll_up", "pause", "jump"],
            weights=[0.7, 0.1, 0.1, 0.1],
        )[0]

        current_position = self.get_scroll_position()

        if action == "scroll_down":
            scroll_length = random.randint(
                self.base_config.download.min_scroll_length,
                self.base_config.download.max_scroll_length,
            )
            target_position = current_position + scroll_length
            self.logger.debug("Trying to scroll down %d pixels", scroll_length)
            actual_position = self.safe_scroll(target_position)
            self.logger.debug("Actually scrolled to %d pixels", actual_position)
            time.sleep(random.uniform(*BaseBehavior.pause_time))
        elif action == "scroll_up":
            scroll_length = random.randint(
                self.base_config.download.min_scroll_length,
                self.base_config.download.max_scroll_length,
            )
            target_position = max(0, current_position - scroll_length)
            self.logger.debug("Trying to scroll up %d pixels", scroll_length)
            actual_position = self.safe_scroll(target_position)
            self.logger.debug("Actually scrolled to %d pixels", actual_position)
        elif action == "pause":
            pause_time = random.uniform(1, 3)
            self.logger.debug("Pausing for %.2f seconds", pause_time)
            time.sleep(pause_time)
        elif action == "jump":
            jump_position = current_position + random.randint(100, 500)
            self.logger.debug("Trying to jump to position %d", jump_position)
            actual_position = self.safe_scroll(jump_position)
            self.logger.debug("Actually jumped to %d", actual_position)

    def safe_scroll(self, target_position: float) -> float:
        current_position = self.get_scroll_position()
        step = random.uniform(
            self.base_config.download.min_scroll_step,
            self.base_config.download.max_scroll_step,
        )
        # step = 50 if target_position > current_position else -50
        # while abs(current_position - target_position) > abs(step):

        while abs(current_position - target_position) > step:
            self.driver.execute_script(f"window.scrollTo(0, {current_position + step});")
            time.sleep(random.uniform(0.01, 0.2))
            new_position = self.get_scroll_position()
            if new_position == current_position:
                self.logger.debug(
                    "Cannot continue scrolling, target: %d, current: %d",
                    target_position,
                    current_position,
                )
                break
            current_position = new_position
        self.driver.execute_script(f"window.scrollTo(0, {target_position});")
        return self.get_scroll_position()

    def get_scroll_position(self) -> float:
        return self.driver.execute_script("return window.pageYOffset")

    def get_page_height(self) -> float:
        return self.driver.execute_script("return document.body.scrollHeight")

    def wait_for_content_load(self) -> None:
        try:
            WebDriverWait(self.driver, 5).until(
                lambda d: d.execute_script("return document.readyState") == "complete",
            )
        except TimeoutException:
            self.logger.warning("Timeout waiting for new content to load")



# v2dl/utils/__init__.py
from .download import AlbumTracker, BaseDownloadAPI, DownloadPathTool, ImageDownloadAPI
from .factory import DownloadAPIFactory, ServiceType, TaskServiceFactory
from .multitask import (
    AsyncService,
    BaseTaskService,
    Task,
    ThreadingService,
)
from .parser import LinkParser
from .security import AccountManager, Encryptor, KeyManager, SecureFileHandler

# only import __all__ when using from automation import *
__all__ = [
    "AccountManager",
    "AlbumTracker",
    "AsyncService",
    "BaseDownloadAPI",
    "BaseTaskService",
    "DownloadAPIFactory",
    "DownloadPathTool",
    "Encryptor",
    "ImageDownloadAPI",
    "KeyManager",
    "LinkParser",
    "SecureFileHandler",
    "ServiceType",
    "Task",
    "TaskServiceFactory",
    "ThreadingService",
]
